if ($host.Version.Major -LT 5) {
	throw "PowerShell version outdated. Please update. Required version: 5.1.x.x."
}

if ($host.Version.Minor -LT 1) {
	if ($host.Version.Major -LE 5) {
		throw "PowerShell version outdated. Please update. Required version: 5.1.x.x."
	}
}

Add-Type @"
using System;
using System.Net;
using System.Net.Security;
using System.Security.Cryptography.X509Certificates;
public class ServerCertificateValidationCallback {
    public static void DoNotIgnore(Object[] certificates) {
		ServicePointManager.ServerCertificateValidationCallback += 
			delegate(Object sender, X509Certificate certificate, X509Chain chain, SslPolicyErrors errors) {
                foreach(Object trustedCertificate in certificates) {
                    if (trustedCertificate.Equals(certificate)) {
                        return true;
                    }
                }
				return false;
			};
	}

	public static void Ignore() {
		ServicePointManager.ServerCertificateValidationCallback += 
			delegate(Object sender, X509Certificate certificate, X509Chain chain, SslPolicyErrors errors) {
				return true;
			};
	}

	public static void AcceptOnly(X509Certificate cert, Object[] certificates) {
		ServicePointManager.ServerCertificateValidationCallback += 
			delegate(Object sender, X509Certificate certificate, X509Chain chain, SslPolicyErrors errors) {
				if (cert.Equals(certificate)) {
                    return true;
                }

                foreach(Object trustedCertificate in certificates) {
                    if (trustedCertificate.Equals(certificate)) {
                        return true;
                    }
                }

                return false;
			};
	}
}
"@

$script:Version='4.10.16'
$script:APIVersion=0
$script:Address=''
$script:Port=443
$script:Credential=$null
$script:CertificateCheckDisabled=$false

# Disable Certificate check (at your own risk!)
# This will not only disable the certificate check for anynode requests, but for
# all requests.
function Disable-CertificateCheck {
    $script:CertificateCheckDisabled=$true
	[ServerCertificateValidationCallback]::Ignore()
}

function Enable-CertificateCheck {
    $script:CertificateCheckDisabled=$false
	[ServerCertificateValidationCallback]::DoNotIgnore((Get-ChildItem 'Cert:/CurrentUser/Root/'))
}

function Initialize-anynodeAddress {
    [CmdletBinding()]
    Param(
        [Parameter(mandatory=$true)]
        [string] $Address
    )

    if ($Address -eq $null -OR $Address -eq '') {
        echo 'Please enter an address.'
        return
    }

    $script:Address=$Address
    echo "Address to communicate with anynode is set to: $Address"
}

function Reset-anynodeAddress {
    $script:Address=''
    echo 'Reset address used to communicate with anynode.'
}

function Initialize-anynodePort {
    [CmdletBinding()]
    Param(
        [Parameter(mandatory=$true)]
        [string] $Port
    )

    if ($Port -eq $null -OR $Port -le 0) {
        echo 'Please enter a valid port.'
        return
    }

    $script:Port=$Port
    echo "Port to communicate with anynode is set to: $Port"
}

function Reset-anynodePort {
    $script:Port=443
    echo 'Reset port used to communicate with anynode to value: 443'
}

function Initialize-anynodeCredential {
    [CmdletBinding(DefaultParameterSetName='cred')]
    Param(
        [Parameter(mandatory=$true,ParameterSetName='cred')]
        [PSCredential] $Credential,
        [Parameter(mandatory=$true,ParameterSetName='userpass')]
        [string] $Username,
        [Parameter(mandatory=$true,ParameterSetName='userpass')]
        [string] $Password
    )

    if ($Credential -ne $null) {
        $script:Credential=$Credential
    } elseif ($Username -ne $null) {
        try {
            $SecretPassword=ConvertTo-SecureString $Password -AsPlainText -Force
	        $script:Credential=New-Object System.Management.Automation.PSCredential($Username,$SecretPassword)
        } catch {
            echo 'Could not set default credential.'
            return
        }
    }

    echo 'Set default credential.'
}

function Reset-anynodeCredential {
    $script:Credential=$null
    echo 'Reset default credential.'
}

# During this session, you can only communicate with the given interface.
function Get-anynodeWebserverCertificate {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $CertificateFilePath
	)

	if ($Address -eq '' -OR $Address -eq $null) {
        if ($script:Address -eq '') {
            $Address=Read-Host 'Address'
        } else {
            $Address=$script:Address
        }

        if ($Address -eq '') {
            echo 'Address required.'
            return
        }
	}

	if ($Port -eq 0) {
		if ($script:Port -eq 0) {
            $Port=Read-Host 'Port'
        } else {
            $Port=$script:Port
        }

        if ($Port -eq 0) {
            echo 'Port required.'
            return
        }
	}

    Disable-CertificateCheck

	try {
		$Request=[System.Net.WebRequest]::Create('https://'+$Address+':'+$Port+'/')
		$Request.GetResponse().Dispose()
		$Certificate=$Request.ServicePoint.Certificate

		if (!($CertificateFilePath -eq $null) -AND !($CertificateFilePath -eq '')) {
			$CertBytes=$Certificate.Export([System.Security.Cryptography.X509Certificates.X509ContentType]::Cert)
			Set-Content -Value $CertBytes -Encoding Byte -Path $CertificateFilePath
		}

		[ServerCertificateValidationCallback]::AcceptOnly($Certificate, (Get-ChildItem 'Cert:/CurrentUser/Root/'))
	} catch {
        Enable-CertificateCheck
		$_.Exception
		return
	}
}

function Import-anynodeWebserverCertificate {
	[CmdletBinding()]
	Param(
		[Parameter(mandatory=$true)]
		[string] $CertificatePath
	)

	try {
		$Certificate=[X509Certificate]::CreateFromCertFile($CertificatePath)
	} catch {
		$_.Exception
		return
	}

	[ServerCertificateValidationCallback]::AcceptOnly($Certificate, (Get-ChildItem 'Cert:/CurrentUser/Root/'))
    echo "Set certificate to $CertificatePath"
}

# Invokes
#	Download
function Invoke-anynodeDownload {
	[CmdletBinding()]
	Param(		
		[Parameter()]
		[ValidateSet('Get','Put','Post','Delete')]
		[string] $Method="GET",
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $URITail,
		[Parameter()]
		[PsObject] $Body,
		[Parameter()]
		[string] $Filename,
		[Parameter()]
		[Hashtable] $Parameters,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	if ($APIVersion -eq 0) {
		$APIVersion=$script:APIVersion
	}

	if ($Address -eq '' -OR $Address -eq $null) {
        if ($script:Address -eq '') {
            $Address=Read-Host 'Address'
        } else {
            $Address=$script:Address
        }

        if ($Address -eq '') {
            echo 'Address required.'
            return
        }
	}

	
	if ($Method -eq '' -OR $Method -eq $null) {
		$Method = "GET"
	}

	if ($Port -eq 0) {
		if ($script:Port -eq 0) {
            $Port=Read-Host 'Port'
        } else {
            $Port=$script:Port
        }

        if ($Port -eq 0) {
            echo 'Port required.'
            return
        }
	}

	if (!($Credential -eq $null)) {
		if (!($Username -eq '') -Or !($Password -eq '')) {
			echo 'Use either Credentials OR Username and Password. Not both.'
			return
		}
	}

	if (!($Username -eq '')) {
		if ($Password -eq '') {
			echo 'To use Username and Password, both must be given.'
			return
		}

		$SecretPassword=ConvertTo-SecureString $Password -AsPlainText -Force
		$Credential=New-Object System.Management.Automation.PSCredential($Username,$SecretPassword)
	}

    if ($Credential -eq $null -AND $script:Credential -ne $null) {
        $Credential=$script:Credential
    }

	$URI='https://'+$Address+':'+$Port+'/api'

	$ParameterString='?version='+$APIVersion+'&filename='+$Filename
	foreach ($Parameter in $Parameters.Keys) {
		$ParameterString+='&'+$Parameter+'='+$Parameters.$Parameter
	}

	if (!($BackendName -eq '')) {
		$ParameterString+='&backendName='+$BackendName
	}

	if (!($Credential -eq $null)) {
		$Username=$Credential.Username
		$Password=$Credential.GetNetworkCredential().Password

		$Base64String=[System.Convert]::ToBase64String([Text.Encoding]::ASCII.GetBytes($Username+':'+$Password))
	}

	if (!($Base64String -eq $null)) {
		$Headers=@{Authorization="Basic "+$Base64String}
	}

    $Base64String=$null

	$URI='https://'+$Address+':'+$Port+'/api'+$URITail+$ParameterString

	try {
        if (($Method -eq 'PUT') -Or ($Method -eq 'Put') -Or ($Method -eq 'put') -Or ($Method -eq 'POST') -Or ($Method -eq 'Post') -Or ($Method -eq 'post')) {

			$Request=Invoke-RestMethod -Method $Method -ContentType 'application/json' -URI $URI -Headers $Headers -Body $Body -OutFile $Filename

		} else {

		    $Request=Invoke-RestMethod -Method $Method -URI $URI -Headers $Headers -OutFile $Filename

        }

		$Request
	} catch {
        try {
            if ($script:CertificateCheckDisabled) {
                $URI='http://'+$Address+':'+$Port+'/api'+$URITail+$ParameterString
                if (($Method -eq 'PUT') -Or ($Method -eq 'Put') -Or ($Method -eq 'put') -Or ($Method -eq 'POST') -Or ($Method -eq 'Post') -Or ($Method -eq 'post')) {

		    	    $Request=Invoke-RestMethod -Method $Method -ContentType 'application/json' -URI $URI -Headers $Headers -Body $Body -OutFile $Filename
        
		        } else {

		            $Request=Invoke-RestMethod -Method $Method -URI $URI -Headers $Headers -OutFile $Filename

                }

		        $Request
            } else {
                throw $_
            }
        } catch {
            $e=$_.Exception
		    $output=$e.InnerException

		    if ($output.Length -eq 0) {
			    $output=$e.Message+' '+$_.ToString()
		    }

		    $output
			return
        }
	}

	if ($Filename -ne $null -AND $Filename -ne '') {
		echo "Saved file at `"$Filename`"."
		return
	}
}

#	Request
function Invoke-anynodeRequest {
	[CmdletBinding()]
	Param(
		[Parameter(mandatory=$true)]
		[ValidateSet('Get','Put','Post','Delete')]
		[string] $Method,
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $URITail,
		[Parameter()]
		[string] $Body,
		[Parameter()]
		[Hashtable] $Parameters,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	if ($APIVersion -eq 0) {
		$APIVersion=$script:APIVersion
	}

	if ($Address -eq '' -OR $Address -eq $null) {
        if ($script:Address -eq '') {
            $Address=Read-Host 'Address'
        } else {
            $Address=$script:Address
        }

        if ($Address -eq '') {
            echo 'Address required.'
            return
        }
	}

	if ($Port -eq 0) {
		if ($script:Port -eq 0) {
            $Port=Read-Host 'Port'
        } else {
            $Port=$script:Port
        }

        if ($Port -eq 0) {
            echo 'Port required.'
            return
        }
	}

	if (!($Credential -eq $null)) {
		if (!($Username -eq '') -Or !($Password -eq '')) {
			echo 'Use either Credentials OR Username and Password. Not both.'
			return
		}
	}

	if (!($Username -eq '')) {
		if ($Password -eq '') {
			echo 'To use Username and Password, both must be given.'
			return
		}

		$SecretPassword=ConvertTo-SecureString $Password -AsPlainText -Force
		$Credential=New-Object System.Management.Automation.PSCredential($Username,$SecretPassword)
	}

    if ($Credential -eq $null -AND $script:Credential -ne $null) {
        $Credential=$script:Credential
    }

	$URI='https://'+$Address+':'+$Port+'/api'

	$ParameterString='?version='+$APIVersion
	foreach ($Parameter in $Parameters.Keys) {
		$ParameterString+='&'+$Parameter+'='+$Parameters.$Parameter
	}

	if (!($BackendName -eq '')) {
		$ParameterString+='&backendName='+$BackendName
	}

	if (!($Credential -eq $null)) {
		$Username=$Credential.Username
		$Password=$Credential.GetNetworkCredential().Password

		$Base64String=[System.Convert]::ToBase64String([Text.Encoding]::ASCII.GetBytes($Username+':'+$Password))
	}

	if (!($Base64String -eq $null)) {
		$Headers=@{Authorization="Basic "+$Base64String}
	}

    $Base64String=$null

	$URI='https://'+$Address+':'+$Port+'/api'+$URITail+$ParameterString

	try {
		if (($Method -eq 'PUT') -Or ($Method -eq 'Put') -Or ($Method -eq 'put') -Or ($Method -eq 'POST') -Or ($Method -eq 'Post') -Or ($Method -eq 'post')) {
			$Request=Invoke-RestMethod -Method $Method -ContentType 'application/json' -URI $URI -Headers $Headers -Body $Body

		} else {
			$Request=Invoke-RestMethod -Method $Method -URI $URI -Headers $Headers
		}

		$Request
	} catch {
        try {
            if ($script:CertificateCheckDisabled) {
                $URI='http://'+$Address+':'+$Port+'/api'+$URITail+$ParameterString
                if (($Method -eq 'PUT') -Or ($Method -eq 'Put') -Or ($Method -eq 'put') -Or ($Method -eq 'POST') -Or ($Method -eq 'Post') -Or ($Method -eq 'post')) {
			        $Request=Invoke-RestMethod -Method $Method -ContentType 'application/json' -URI $URI -Headers $Headers -Body $Body

		        } else {
			        $Request=Invoke-RestMethod -Method $Method -URI $URI -Headers $Headers
		        }

		        $Request
				return
            } else {
                throw $_
            }
        } catch {
			$e=$_.Exception
			$output=$e.InnerException

			if ($output.Length -eq 0) {
				$output=$e.Message+' '+$_.ToString()
			}

			$output
        }
	}
}

#	Upload
function Invoke-anynodeUpload {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $URITail,
		[Parameter(mandatory=$true)]
		[Object[]] $Filenames,
		[Parameter()]
		[Hashtable] $Parameters,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	if ($APIVersion -eq 0) {
		$APIVersion=$script:APIVersion
	}

	if ($Address -eq '' -OR $Address -eq $null) {
        if ($script:Address -eq '') {
            $Address=Read-Host 'Address'
        } else {
            $Address=$script:Address
        }

        if ($Address -eq '') {
            echo 'Address required.'
            return
        }
	}

	if ($Port -eq 0) {
		if ($script:Port -eq 0) {
            $Port=Read-Host 'Port'
        } else {
            $Port=$script:Port
        }

        if ($Port -eq 0) {
            echo 'Port required.'
            return
        }
	}

	if (!($Credential -eq $null)) {
		if (!($Username -eq '') -Or !($Password -eq '')) {
			echo 'Use either Credentials OR Username and Password. Not both.'
			return
		}
	}

	if (!($Username -eq '')) {
		if ($Password -eq '') {
			echo 'To use Username and Password, both must be given.'
			return
		}

		$SecretPassword=ConvertTo-SecureString $Password -AsPlainText -Force
		$Credential=New-Object System.Management.Automation.PSCredential($Username,$SecretPassword)
	}

    if ($Credential -eq $null -AND $script:Credential -ne $null) {
        $Credential=$script:Credential
    }

	$URI='https://'+$Address+':'+$Port+'/api'

	$ParameterString='?version='+$APIVersion
	foreach ($Parameter in $Parameters.Keys) {
		$ParameterString+='&'+$Parameter+'='+$Parameters.$Parameter
	}

	if (!($BackendName -eq '')) {
		$ParameterString+='&backendName='+$BackendName
	}

	if (!($Credential -eq $null)) {
		$Username=$Credential.Username
		$Password=$Credential.GetNetworkCredential().Password

		$Base64String=[System.Convert]::ToBase64String([Text.Encoding]::ASCII.GetBytes($Username+':'+$Password))
	}

	$URI='https://'+$Address+':'+$Port+'/api'+$URITail+$ParameterString

	if (!($Base64String -eq $null)) {
		$Headers=@{Authorization="Basic "+$Base64String}
	}

    $Base64String=$null

	$Boundary=[System.Guid]::NewGuid().ToString()
	$ContentType='multipart/form-data; boundary="'+$Boundary+'"'

	$NL="`r`n";

	$Body=[byte[]]::new(0)

	foreach ($Filename in $Filenames) {
		$Bytes=[System.IO.File]::ReadAllBytes($Filename)

		$Head=(
			"--$Boundary",
		    "Content-Disposition: form-data; name=`"file`"; filename=`"$Filename`"",
		    "Content-Type: application/octet-stream$NL$NL"
		) -join $NL

		$Body+=[Text.Encoding]::UTF8.GetBytes($Head)
		$Body+=$Bytes

		$Foot="$NL--$Boundary--$NL"
		$Body+=[Text.Encoding]::UTF8.GetBytes($Foot)
	}

	$tmp=New-TemporaryFile
	$tmpName=$tmp.FullName
	[IO.File]::WriteAllBytes($tmpName,$Body)

	try {
		Invoke-RestMethod -Method PUT -Uri $URI -Headers $Headers -ContentType $ContentType -InFile $tmpName
	} catch {
        try {
            if ($script:CertificateCheckDisabled) {
                $URI='http://'+$Address+':'+$Port+'/api'+$URITail+$ParameterString
                Invoke-RestMethod -Method PUT -Uri $URI -Headers $Headers -ContentType $ContentType -InFile $tmpName

		        $Request
				return
            } else {
                throw $_
            }
        } catch {
        }

		$e=$_.Exception
		$output=$e.InnerException

		if ($output.Length -eq 0) {
			$output=$e.Message+' '+$_.ToString()
		}

		$output
	} finally {
		Remove-Item $tmpName -Force	
	}
}

# Active Sessions
#	Get
function Get-anynodeActiveSessions {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[ValidateSet('ISO-8601','XZWO')]
		[string] $DateFormat,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if ($DateFormat -ne $null -AND $DateFormat -ne '') {
		$Parameters.Add('dateFormat',$DateFormat)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /activeSessions/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Backends
#	Get
function Get-anynodeBackends {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /backends/get -Credential $Credential -APIVersion $APIVersion -Username $Username -Password $Password
}

# Config
#	Get
function Get-anynodeConfig {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $Filename,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeDownload -Address $Address -Port $Port -URITail /anynodeConfig/get -Filename $Filename -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Import
function Import-anynodeConfig {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Filepath,
		[switch] $CommitProtection,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=''
	try {
		$Body=[IO.File]::ReadAllText($Filepath)
	} catch {
		echo 'Could not find file.'
		return
	}

	$Parameters=@{}
	if ($CommitProtection) {
		$Parameters.Add("commitProtection",$true)
	}

	Invoke-anynodeUpload -Address $Address -Port $Port -URITail /anynodeConfig/set -Filenames @($Filepath) -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Push
function Push-anynodeConfig {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /anynodeConfig/commit -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Remove
function Remove-anynodeConfig {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[switch] $Force,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	if (($Force -eq $null) -Or ($Force -eq $false)) {
		$Answer=Read-Host 'Are you sure you want to clean the whole anynode configuration? (y/n)'
	} elseif ($Force -eq $true) {
		$Answer='y'
	}

	if ($Answer -ne 'y') {
		return
	}

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /anynodeConfig/clean -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Reset
function Reset-anynodeConfig {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /anynodeConfig/revert -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Dashboard
#	Get
function Get-anynodeDashboard {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[ValidateSet('sipNodes','sfbUcmaNodes','ldapConnections','certificates','networkControllers','ldapDirectories','numberValidationRouteSupervision','azureActiveDirectories','restClientRouteSupervision','directoryRouteSupervision')]
		[psobject] $List,
		[Parameter()]
		[ValidateSet('ISO-8601','XZWO')]
		[string] $DateFormat,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if (!($List -eq $null)) {
		$Parameters.Add("list",$List)
	}

	if ($DateFormat -ne $null -AND $DateFormat -ne '') {
		$Parameters.Add('dateFormat',$DateFormat)
	}
	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /dashboard/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Dial String Rewriting
#	Get
function Get-anynodeDialStringRewriting {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $DialStringRewriting,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)
	
	$Parameters=@{}
	if ($DialStringRewriting -ne $null) {
		$Parameters.Add("dialStringRewriting",$DialStringRewriting)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /dialStringRewriting/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set
function Set-anynodeDialStringRewriting {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $DialStringRewriting,
		[Parameter(mandatory=$true)]
		[string] $DisplayName,
		[Parameter()]
		[string] $Id,
		[Parameter()]
		[psobject] $Rules,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)
	
	$Parameters=@{}
	if ($DialStringRewriting -ne $null) {
		$Parameters.Add("dialStringRewriting",$DialStringRewriting)
	}
	
	$Body=@{}
	if ($DisplayName -ne $null) {
		$Body.Add("displayName",$DisplayName)
	}
	
	if ($Id -ne $null) {
		$Body.Add("id",$Id)
	}

	if ($Rules -ne $null) {
		$Body.Add("rules",$Rules)
	}

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /dialStringRewriting/set -Parameters $Parameters -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add
function Add-anynodeDialStringRewriting {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $DisplayName,
		[Parameter()]
		[string] $Id,
		[Parameter()]
		[psobject] $Rules,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)
	
	$Body=@{}
	if ($DisplayName -ne $null) {
		$Body.Add("displayName",$DisplayName)
	}
	
	if ($Id -ne $null) {
		$Body.Add("id",$Id)
	}

	if ($Rules -ne $null) {
		$Body.Add("rules",$Rules)
	}

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /dialStringRewriting/add -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Remove
function Remove-anynodeDialStringRewriting {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $DialStringRewriting,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)
	
	$Parameters=@{}
	if ($DialStringRewriting -ne $null) {
		$Parameters.Add("dialStringRewriting",$DialStringRewriting)
	}

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /dialStringRewriting/remove -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get Rules
function Get-anynodeDialStringRewritingRules {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $DialStringRewriting,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)
	
	$Parameters=@{}
	if ($DialStringRewriting -ne $null) {
		$Parameters.Add("dialStringRewriting",$DialStringRewriting)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /dialStringRewriting/rules/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set Rules
function Set-anynodeDialStringRewritingRules {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $DialStringRewriting,
		[Parameter(mandatory=$true)]
		[psobject] $Rules,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)
	
	$Parameters=@{}
	if ($DialStringRewriting -ne $null) {
		$Parameters.Add("dialStringRewriting",$DialStringRewriting)
	}

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /dialStringRewriting/rules/set -Parameters $Parameters -Body ($Rules | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Import Rules from CSV
function Import-anynodeDialStringRewritingRules {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $DialStringRewriting,
		[Parameter(mandatory=$true)]
		[string] $Filename,
		[Parameter()]
		[int] $ColumnIndex,
		[Parameter()]
		[ValidateSet($true,$false)]
		[string] $CutExtensionNumber,
		[ValidateSet($true,$false)]
		[string] $ClearDisplayName,
		[ValidateSet($true,$false)]
		[string] $ProcessingDone,
		[ValidateSet($true,$false)]
		[string] $Append,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if ($DialStringRewriting -ne $null) {
		$Parameters.Add("dialStringRewriting",$DialStringRewriting)
	}

	if ($ColumnIndex -ne $null) {
		$Parameters.Add("columnIndex",$ColumnIndex)
	}

	if ($CutExtensionNumber -ne $null) {
		$Parameters.Add("cutExtensionNumber",$CutExtensionNumber)
	}

	if ($ClearDisplayName -ne $null) {
		$Parameters.Add("clearDisplayName",$ClearDisplayName)
	}

	if ($ProcessingDone -ne $null) {
		$Parameters.Add("processingDone",$ProcessingDone)
	}

	if ($Append -ne $null) {
		$Parameters.Add("append",$Append)
	}

	Invoke-anynodeUpload -Address $Address -Port $Port -URITail /dialStringRewriting/rules/import -Filenames $Filename -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Export Rules as JSON
function Export-anynodeDialStringRewritingRules {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $DialStringRewriting,
		[Parameter(mandatory=$true)]
		[string] $Filename,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if ($DialStringRewriting -ne $null) {
		$Parameters.Add("dialStringRewriting",$DialStringRewriting)
	}

	Invoke-anynodeDownload -Address $Address -Port $Port -URITail /dialStringRewriting/rules/export -Filename $Filename -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add Rule
function Add-anynodeDialStringRewritingRule {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $DialStringRewriting,
		[Parameter()]
		[int] $Index,
		[Parameter()]
		[ValidateSet('prefixAndSuffix','wildcardPattern','cutExtensionNumber','matchAndModify','matchAndBranch')]
		[string] $RewriteType,
		[Parameter()]
		[string] $Prefix,
		[Parameter()]
		[string] $Suffix,
		[Parameter()]
		[string] $WildcardPattern,
		[Parameter()]
		[int] $DeleteLeadingCharacters,
		[Parameter()]
		[int] $DeleteTrailingCharacters,
		[Parameter()]
		[string] $AddPrefix,
		[Parameter()]
		[string] $AddSuffix,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $ClearDisplayName,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[ValidateSet('lowercase','uppercase','casefold')]
		[string] $Case,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $ApplyToAllDialStrings,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $ApplyToSourceDialString,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $ApplyToDestinationDialString,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $SkipFurtherRules,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $SkipRulesBelow,
		[Parameter()]
		[string] $Comment,
		[Parameter()]
		#Valid list of ('unspecified','sessionSource','sessionAsserted','sessionDestination',
		#				'sessionConnected','sessionElin','sessionRedirectHistory',
		#				'sessionTransferrer','reasonRedirection','reasonSelected',
		#				'mwiSource','mwiDestination')
		[psobject] $ApplyTo,
		[Parameter()]
		#This is a list of strings
		[psobject] $MatchDialStringList,
		[Parameter()]
		#Values are ('trunk','firstExtension','lastExtension','zeroExtend')
		[psobject] $MatchExtensionRange,
		[Parameter()]
		#List of object containing ('comment','type','digits','text','minCount',
		#							'maxCount','greedyMatch')
		[psobject] $StructuralPattern,
		[Parameter()]
		#List of TelRewriteSegment
		[psobject] $RewriteSegments,
		[Parameter()]
		#List of this object
		[psobject] $BranchRewrites,
		[Parameter()]
		#List of this object
		[psobject] $BranchElseRewrites,
		[Parameter()]
		#Valid list of ('or','xor','and','not')
		[psobject] $MatchTags,
		[Parameter()]
		#Values are ('clear','operations')
		[psobject] $TagRewrite,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if ($DialStringRewriting -ne $null) {
		$Parameters.Add("dialStringRewriting",$DialStringRewriting)
	}

	$Body=@{}
	if ($Index -ne $null) {
		$Body.Add("index",$Index)
	}
	
	if ($RewriteType -ne $null) {
		$Body.Add("rewriteType",$RewriteType)
	}

	if ($Prefix -ne $null) {
		$Body.Add("prefix",$Prefix)
	}

	if ($Suffix -ne $null) {
		$Body.Add("suffix",$Suffix)
	}

	if ($WildcardPattern -ne $null) {
		$Body.Add("wildcardPattern",$WildcardPattern)
	}

	if ($DeleteLeadingCharacters -ne $null) {
		$Body.Add("deleteLeadingCharacters",$DeleteLeadingCharacters)
	}

	if ($DeleteTrailingCharacters -ne $null) {
		$Body.Add("deleteTrailingCharacters",$DeleteTrailingCharacters)
	}

	if ($AddPrefix -ne $null) {
		$Body.Add("addPrefix",$AddPrefix)
	}

	if ($AddSuffix -ne $null) {
		$Body.Add("addSuffix",$AddSuffix)
	}

	if ($ClearDisplayName -ne $null) {
		$Body.Add("clearDisplayName",$ClearDisplayName)
	}

	if ($DisplayName -ne $null) {
		$Body.Add("displayName",$DisplayName)
	}

	if ($Case -ne $null) {
		$Body.Add("case",$Case)
	}

	if ($ApplyToAllDialStrings -ne $null) {
		$Body.Add("applyToAllDialStrings",$ApplyToAllDialStrings)
	}

	if ($ApplyToSourceDialString -ne $null) {
		$Body.Add("applyToSourceDialString",$ApplyToSourceDialString)
	}

	if ($ApplyToDestinationDialString -ne $null) {
		$Body.Add("applyToDestinationDialString",$ApplyToDestinationDialString)
	}

	if ($SkipFurtherRules -ne $null) {
		$Body.Add("skipFurtherRules",$SkipFurtherRules)
	}

	if ($SkipRulesBelow -ne $null) {
		$Body.Add("skipRulesBelow",$SkipRulesBelow)
	}

	if ($Comment -ne $null) {
		$Body.Add("comment",$Comment)
	}

	if ($ApplyTo -ne $null) {
		$Body.Add("applyTo",$ApplyTo)
	}

	if ($MatchDialStringList -ne $null) {
		$Body.Add("matchDialStringList",$MatchDialStringList)
	}

	if ($MatchExtensionRange -ne $null) {
		$Body.Add("matchExtensionRange",$MatchExtensionRange)
	}

	if ($StructuralPattern -ne $null) {
		$Body.Add("structuralPattern",$StructuralPattern)
	}

	if ($RewriteSegments -ne $null) {
		$Body.Add("rewriteSegments",$RewriteSegments)
	}

	if ($BranchRewrites -ne $null) {
		$Body.Add("branchRewrites",$BranchRewrites)
	}

	if ($BranchElseRewrites -ne $null) {
		$Body.Add("branchElseRewrites",$BranchElseRewrites)
	}

	if ($MatchTags -ne $null) {
		$Body.Add("matchTags",$MatchTags)
	}

	if ($TagRewrite -ne $null) {
		$Body.Add("tagRewrite",$TagRewrite)
	}

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /dialStringRewriting/rules/add -Parameters $Parameters -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Remove Rule
function Remove-anynodeDialStringRewritingRule {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $DialStringRewriting,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[int] $Index,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if ($DialStringRewriting -ne $null) {
		$Parameters.Add("dialStringRewriting",$DialStringRewriting)
	}

	if ($DisplayName -ne $null) {
		$Parameters.Add("displayName",$DisplayName)
	}

	if ($Index -ne $null) {
		$Parameters.Add("index",$Index)
	}

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /dialStringRewriting/rules/remove -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Clear Rules
function Clear-anynodeDialStringRewritingRules {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $DialStringRewriting,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if ($DialStringRewriting -ne $null) {
		$Parameters.Add("dialStringRewriting",$DialStringRewriting)
	}

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /dialStringRewriting/rules/clear -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get
function Get-anynodeDialStringRewritings {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /dialStringRewriting/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Forwarded Session Recording
#	Search
function Search-anynodeForwardedSessionRecordingIdentifier {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $SipCallId,
		[Parameter()]
		[string] $SipLocalTag,
		[Parameter()]
		[string] $SipRemoteTag,
		[Parameter()]
		[string] $SipUserAgentDialog,
		[Parameter()]
		[string] $SdpOrigin,
		[Parameter()]
		[string] $SdpMediaLineAddress,
		[Parameter()]
		[string] $SdpMediaLinePort,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$ParamPresent=$false

	$Body=@{}
	if ($SipCallId -ne '') {
		$Body.Add('sipCallId',$SipCallId)
		$ParamPresent=$true
	}

	if ($SipLocalTag -ne '') {
		$Body.Add('sipLocalTag',$SipLocalTag)
		$ParamPresent=$true
	}

	if ($SipRemoteTag -ne '') {
		$Body.Add('sipRemoteTag',$SipRemoteTag)
		$ParamPresent=$true
	}

	if ($SipUserAgentDialog -ne '') {
		$Body.Add('sipUserAgentDialog',$SipUserAgentDialog)
		$ParamPresent=$true
	}

	if ($SdpOrigin -ne '') {
		$Body.Add('sdpOrigin',$SdpOrigin)
		$ParamPresent=$true
	}

	if ($SdpMediaLineAddress -ne '') {
		$Body.Add('sdpMLineAddress',$SdpMediaLineAddress)
		$ParamPresent=$true
	}

	if ($SdpMediaLinePort -ne '') {
		$Body.Add('sdpMLinePort',$SdpMediaLinePort)
		$ParamPresent=$true
	}

	if (!$ParamPresent) {
		echo 'At least one filter parameter has to be given.'
		return
	}

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /forwardedSessionRecording/lookupByLocalSessionIdentification -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set Mute
function Set-anynodeForwardedSessionRecordingMute {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Identifier,
		[Parameter(mandatory=$true)]
		[ValidateSet($true,$false)]
		[bool] $Mute,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}
	$Body.Add('identifier',$Identifier)

	$Body.Add('mute',$Mute)

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /forwardedSessionRecording/setMute -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# License
#	Get
function Get-anynodeLicenses {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[ValidateSet('ISO-8601','XZWO')]
		[string] $DateFormat,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if ($DateFormat -ne $null -AND $DateFormat -ne '') {
		$Parameters.Add('dateFormat',$DateFormat)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /licenses/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Clear
function Clear-anynodeLicenses {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /licenses/clear -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Import
function Import-anynodeLicense {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Filepath,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=''
	try {
		$Body=[IO.File]::ReadAllText($Filepath)
	} catch {
		echo 'Could not find file.'
		return
	}

	Invoke-anynodeUpload -Address $Address -Port $Port -URITail /licenses/importFile -Filenames @($Filepath) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# LOD
#	Import
function Import-anynodeLOD {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $LODKey,
		[Parameter()]
		[string] $LicensingDevice,
		[Parameter(mandatory=$true)]
		[string] $EmailAddress,
		[Parameter(mandatory=$true)]
		[string] $CompanyName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}

	$Body.Add("lodKey",$LODKey)

	if (!($LicensingDevice -eq '')) {
		$Body.Add("licensingDevice",$LicensingDevice)
	}

	$Body.Add("emailAddress",$EmailAddress)

	$Body.Add("companyName",$CompanyName)

	$license=Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /licenses/importLod -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password

	$Identifier=$license.identifier
	
	if ($Identifier -eq $null) {
		$license
		return
	}

	$LicensesLoaded=Get-anynodeLicenses -Address $Address -Port $Port -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password

	$Result

	foreach ($LicenseLoaded in $LicensesLoaded) {
		if ($LicenseLoaded.identifier -eq $Identifier) {
			$Result=$LicenseLoaded
		}
	}

	if ($Result -eq $null) {
		$license
	} else {
		$Result
	}
}

# Maintenance Mode
#	Get
function Get-anynodeMaintenanceMode {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /maintenanceMode/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set
function Set-anynodeMaintenanceMode {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $ModeActive,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}
	
	if (!($ModeActive -eq $null)) {
		$Body.Add("modeActive",$ModeActive)
	}

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /maintenanceMode/set -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Media Recording
#	Get
function Get-anynodeMediaRecording {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /mediaRecording/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}
# Module Version
#	Get
function Get-anynodeModuleVersion {
	$script:Version
}

# Network Controller
#	Get
function Get-anynodeNetworkControllers {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /networkControllers/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get Interfaces
function Get-anynodeNetworkInterfaces {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /networkInterfaces/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Node
#	Get
function Get-anynodeNodes {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $IncludeNetworkInformation,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

    $Parameters=@{}

    if ($IncludeNetworkInformation -ne $null) {
        $Parameters.Add("includeNetworkInformation",$IncludeNetworkInformation)
    }

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get Provider Profiles
function Get-anynodeProviderNodeProfiles {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[switch] $CompleteData,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if ($CompleteData) {
		$Parameters.Add("completeData",$true)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/providerProfiles/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add Provider - deprecated parameters incomingManipulations and outgoingManipulations
function Add-anynodeProviderNode {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Profile,
		[Parameter()]
		[psobject] $DeutscheTelekomNumberExtensionRange,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $CreateNewDeutscheTelekomSipTrunk=$true,
		[Parameter()]
		[string] $DeutscheTelekomSipTrunkId,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $CreateNewNetworkController=$true,
		[Parameter()]
		[string] $NetworkControllerDisplayName,
		[Parameter()]
		[string] $NetworkControllerInterface='[Any Interface]',
		[Parameter()]
		[ValidateSet('[Any Address Version]','IN_ADDRESS_VERSION_4','IN_ADDRESS_VERSION_6')]
		[string] $NetworkControllerIpVersion='[Any Address Version]',
		[Parameter()]
		[string] $NetworkControllerIpAddress='[Any IP Address]',
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $EnableReverseDnsLookup,
		[Parameter()]
		[string] $NetworkControllerId,
		[Parameter()]
		[int] $UdpTcpPort,
		[Parameter()]
		[int] $TlsPort,
		[Parameter()]
		[psobject] $UdpPortRange,
		[Parameter()]
		[psobject] $TcpPortRange,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $ActivateNatTraversal,
		[Parameter()]
		[string] $ExternalHost,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $NatDeviceHasFixedPortMapping,
		[Parameter()]
		[int] $FixedUdpTcpPort,
		[Parameter()]
		[int] $FixedTlsPort,
		[Parameter()]
		[int] $FixedUdpPortRangeStart,
		[Parameter()]
		[int] $FixedTcpPortRangeStart,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $NodeViaSipTrunking,
		[Parameter()]
		[psobject] $RemoteSipDomain,
		[Parameter()]
		[string] $LocalSipDomain,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $ActivateClientAuthentication,
		[Parameter()]
		[psobject] $ClientCredentials,
		[Parameter()]
		[string] $RegistrarUri,
		[Parameter()]
		[string] $AddressOfRecord,
		[Parameter()]
		[string] $ProxyUri,
		[Parameter()]
		[string] $PrimaryProxyUri,
		[Parameter()]
		[string] $SecondaryProxyUri,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $DefineAssertedUri,
		[Parameter()]
		[string] $AssertedUri,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $UseNetworkPeerWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludeRemoteSipDomainInWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludeSipRegistrarInWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludeProxyInWhitelist,
		[Parameter()]
		[psobject] $NetworkPeerWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $AllowOnlyNegotiatedPeersForRtpRtcp,
        [Parameter()]
		[psobject] $IncomingDialStringRewriteRules,
		[Parameter()]
		[psobject] $OutgoingDialStringRewriteRules,
		[Parameter()]
		[psobject] $RoutingDomains,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion=2,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password,
		[Parameter()]
		[psobject] $IncomingManipulations,
		[Parameter()]
		[psobject] $OutgoingManipulations
	)

	$Parameters=@{profile=$Profile}

	$Body=@{}

	if (!($DeutscheTelekomNumberExtensionRange -eq $null)) {
		$Body.Add("deutscheTelekomNumberExtensionRange",$DeutscheTelekomNumberExtensionRange)
	}

	$Body.Add("createNewDeutscheTelekomSipTrunk",$CreateNewDeutscheTelekomSipTrunk)

	if (!($DeutscheTelekomSipTrunkId -eq $null) -AND !($DeutscheTelekomSipTrunkId -eq '')) {
		$Body.Add("deutscheTelekomSipTrunkId",$DeutscheTelekomSipTrunkId)
	}

	$Body.Add("createNewNetworkController",$CreateNewNetworkController)
	if ($CreateNewNetworkController) {
		if (!($NetworkControllerDisplayName -eq $null) -AND !($NetworkControllerDisplayName -eq '')) {
			$Body.Add("networkControllerDisplayName",$NetworkControllerDisplayName)
		}
		$Body.Add("networkControllerInterface",$NetworkControllerInterface)
		$Body.Add("networkControllerIpVersion",$NetworkControllerIpVersion)
		$Body.Add("networkControllerIpAddress",$NetworkControllerIpAddress)
		if (!($EnableReverseDnsLookup -eq $null)) {
			$Body.Add("enableReverseDnsLookup",$EnableReverseDnsLookup)
		}
	}

	if (!($NetworkControllerId -eq $null) -AND !($NetworkControllerId -eq '')) {
		$Body.Add("networkControllerId",$NetworkControllerId)
	}

	if (!($UdpTcpPort -eq $null) -AND $UdpTcpPort -gt 0) {
		$Body.Add("udpTcpPort",$UdpTcpPort)
	}

	if (!($TlsPort -eq $null) -AND $TlsPort -gt 0) {
		$Body.Add("tlsPort",$TlsPort)
	}

	if (!($UdpPortRange -eq $null)) {
		$Body.Add("udpPortRange",$UdpPortRange)
	}

	if (!($TcpPortRange -eq $null)) {
		$Body.Add("tcpPortRange",$TcpPortRange)
	}

	if (!($ActivateNatTraversal -eq $null)) {
		$Body.Add("activateNatTraversal",$ActivateNatTraversal)
	}
	if ($ActivateNatTraversal) {
		$Body.Add("externalHost",$ExternalHost)

		if (!($NatDeviceHasFixedPortMapping -eq $null)) {
			$Body.Add("natDeviceHasFixedPortMapping",$NatDeviceHasFixedPortMapping)
		}
		if ($NatDeviceHasFixedPortMapping) {
			$Body.Add("fixedUdpTcpPort",$FixedUdpTcpPort)
			$Body.Add("fixedTlsPort",$FixedTlsPort)
			$Body.Add("fixedUdpPortRangeStart",$FixedUdpPortRangeStart)
			$Body.Add("fixedTcpPortRangeStart",$FixedTcpPortRangeStart)
		}
	}

	if (!($NodeViaSipTrunking -eq $null)) {
		$Body.Add("nodeViaSipTrunking",$NodeViaSipTrunking)
	}

	if (!($RemoteSipDomain -eq $null)) {
		$Body.Add("remoteSipDomain",$RemoteSipDomain)
	}

	if (!($LocalSipDomain -eq $null) -AND !($LocalSipDomain -eq '')) {
		$Body.Add("localSipDomain",$LocalSipDomain)
	}

	if (!($ActivateClientAuthentication -eq $null)) {
		$Body.Add("activateClientAuthentication",$ActivateClientAuthentication)
	}

	if (!($ClientCredentials -eq $null)) {
		$Body.Add("credentials",$ClientCredentials)
	}

	if (!($RegistrarUri -eq $null) -AND !($RegistrarUri -eq '')) {
		$Body.Add("registrarUri",$RegistrarUri)
	}

	if (!($AddressOfRecord -eq $null)) {
		$Body.Add("addressOfRecord",$AddressOfRecord)
	}

	if (!($ProxyUri -eq $null) -AND !($ProxyUri -eq '')) {
		$Body.Add("proxyUri",$ProxyUri)
	}

	if (!($PrimaryProxyUri -eq $null) -AND !($PrimaryProxyUri -eq '')) {
		$Body.Add("primaryProxyUri",$PrimaryProxyUri)
	}

	if (!($SecondaryProxyUri -eq $null) -AND !($SecondaryProxyUri -eq '')) {
		$Body.Add("secondaryProxyUri",$SecondaryProxyUri)
	}

	if (!($DefineAssertedUri -eq $null)) {
		$Body.Add("defineAssertedUri",$DefineAssertedUri)
	}
	if ($DefineAssertedUri) {
		$Body.Add("assertedUri",$AssertedUri)
	}

	if (!($UseNetworkPeerWhitelist -eq $null)) {
		$Body.Add("useNetworkPeerWhitelist",$UseNetworkPeerWhitelist)
	}

	$Body.Add("includeRemoteSipDomainInWhitelist",$IncludeRemoteSipDomainInWhitelist)
	$Body.Add("includeSipRegistrarInWhitelist",$IncludeSipRegistrarInWhitelist)
	$Body.Add("includeProxyInWhitelist",$IncludeProxyInWhitelist)
	$Body.Add("networkPeerWhitelist",$NetworkPeerWhitelist)
	$Body.Add("allowOnlyNegotiatedPeersForRtpRtcp",$AllowOnlyNegotiatedPeersForRtpRtcp)

	if (!($IncomingManipulations -eq $null)) {
		if ($IncomingManipulations.GetType().Name -eq 'Object[]') {
			$Body.Add("incomingManipulations",$IncomingManipulations)
		} else {
			$Body.Add("incomingManipulations",@{0=$IncomingManipulations}.Values)
		}
	}

	if (!($OutgoingManipulations -eq $null)) {
		if ($OutgoingManipulations.GetType().Name -eq 'Object[]') {
			$Body.Add("outgoingManipulations",$OutgoingManipulations)
		} else {
			$Body.Add("outgoingManipulations",@{0=$OutgoingManipulations}.Values)
		}
	}

	if (!($IncomingDialStringRewriteRules -eq $null)) {
		if ($IncomingDialStringRewriteRules.GetType().Name -eq 'Object[]') {
			$Body.Add("incomingDialStringRewriteRules",$IncomingDialStringRewriteRules)
		} else {
			$Body.Add("incomingDialStringRewriteRules",@{0=$IncomingDialStringRewriteRules}.Values)
		}
	}

	if (!($OutgoingDialStringRewriteRules -eq $null)) {
		if ($OutgoingDialStringRewriteRules.GetType().Name -eq 'Object[]') {
			$Body.Add("outgoingDialStringRewriteRules",$OutgoingDialStringRewriteRules)
		} else {
			$Body.Add("outgoingDialStringRewriteRules",@{0=$OutgoingDialStringRewriteRules}.Values)
		}
	}

	if (!($RoutingDomains -eq $null)) {
		$Body.Add("routingDomains",$RoutingDomains)
	}

	if (!($DisplayName -eq $null) -AND !($DisplayName -eq '')) {
		$Body.Add("displayName",$DisplayName)
	}

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /nodes/addProvider -Body ($Body | ConvertTo-Json -Depth 99) -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	DEPRECATED
function Add-anynodeProviderNodeV2 {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Profile,
		[Parameter()]
		[psobject] $DeutscheTelekomNumberExtensionRange,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $CreateNewDeutscheTelekomSipTrunk=$true,
		[Parameter()]
		[string] $DeutscheTelekomSipTrunkId,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $CreateNewNetworkController=$true,
		[Parameter()]
		[string] $NetworkControllerDisplayName,
		[Parameter()]
		[string] $NetworkControllerInterface='[Any Interface]',
		[Parameter()]
		[ValidateSet('[Any Address Version]','IN_ADDRESS_VERSION_4','IN_ADDRESS_VERSION_6')]
		[string] $NetworkControllerIpVersion='[Any Address Version]',
		[Parameter()]
		[string] $NetworkControllerIpAddress='[Any IP Address]',
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $EnableReverseDnsLookup,
		[Parameter()]
		[string] $NetworkControllerId,
		[Parameter()]
		[int] $UdpTcpPort,
		[Parameter()]
		[int] $TlsPort,
		[Parameter()]
		[psobject] $UdpPortRange,
		[Parameter()]
		[psobject] $TcpPortRange,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $ActivateNatTraversal,
		[Parameter()]
		[string] $ExternalHost,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $NatDeviceHasFixedPortMapping,
		[Parameter()]
		[int] $FixedUdpTcpPort,
		[Parameter()]
		[int] $FixedTlsPort,
		[Parameter()]
		[int] $FixedUdpPortRangeStart,
		[Parameter()]
		[int] $FixedTcpPortRangeStart,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $NodeViaSipTrunking,
		[Parameter()]
		[psobject] $RemoteSipDomain,
		[Parameter()]
		[string] $LocalSipDomain,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $ActivateClientAuthentication,
		[Parameter()]
		[psobject] $ClientCredentials,
		[Parameter()]
		[string] $RegistrarUri,
		[Parameter()]
		[string] $AddressOfRecord,
		[Parameter()]
		[string] $ProxyUri,
		[Parameter()]
		[string] $PrimaryProxyUri,
		[Parameter()]
		[string] $SecondaryProxyUri,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $DefineAssertedUri,
		[Parameter()]
		[string] $AssertedUri,
		[Parameter()]
		[string] $AssertedUriHost,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $UseNetworkPeerWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludeRemoteSipDomainInWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludeSipRegistrarInWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludeProxyInWhitelist,
		[Parameter()]
		[psobject] $NetworkPeerWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $AllowOnlyNegotiatedPeersForRtpRtcp,
        [Parameter()]
		[psobject] $IncomingDialStringRewriteRules,
		[Parameter()]
		[psobject] $OutgoingDialStringRewriteRules,
		[Parameter()]
		[psobject] $RoutingDomains,
        [Parameter()]
		[ValidateSet("SCHWEIZ","LIECHTENSTEIN")]
		[psobject] $Region,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion=2,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{profile=$Profile}

	$Body=@{}

    if(!($Region -eq $null)) {
        $Body.Add("region", $Region)
    }

	if (!($DeutscheTelekomNumberExtensionRange -eq $null)) {
		$Body.Add("deutscheTelekomNumberExtensionRange",$DeutscheTelekomNumberExtensionRange)
	}

	$Body.Add("createNewDeutscheTelekomSipTrunk",$CreateNewDeutscheTelekomSipTrunk)

	if (!($DeutscheTelekomSipTrunkId -eq $null) -AND !($DeutscheTelekomSipTrunkId -eq '')) {
		$Body.Add("deutscheTelekomSipTrunkId",$DeutscheTelekomSipTrunkId)
	}

	$Body.Add("createNewNetworkController",$CreateNewNetworkController)
	if ($CreateNewNetworkController) {
		if (!($NetworkControllerDisplayName -eq $null) -AND !($NetworkControllerDisplayName -eq '')) {
			$Body.Add("networkControllerDisplayName",$NetworkControllerDisplayName)
		}
		$Body.Add("networkControllerInterface",$NetworkControllerInterface)
		$Body.Add("networkControllerIpVersion",$NetworkControllerIpVersion)
		$Body.Add("networkControllerIpAddress",$NetworkControllerIpAddress)
		if (!($EnableReverseDnsLookup -eq $null)) {
			$Body.Add("enableReverseDnsLookup",$EnableReverseDnsLookup)
		}
	}

	if (!($NetworkControllerId -eq $null) -AND !($NetworkControllerId -eq '')) {
		$Body.Add("networkControllerId",$NetworkControllerId)
	}

	if (!($UdpTcpPort -eq $null) -AND $UdpTcpPort -gt 0) {
		$Body.Add("udpTcpPort",$UdpTcpPort)
	}

	if (!($TlsPort -eq $null) -AND $TlsPort -gt 0) {
		$Body.Add("tlsPort",$TlsPort)
	}

	if (!($UdpPortRange -eq $null)) {
		$Body.Add("udpPortRange",$UdpPortRange)
	}

	if (!($TcpPortRange -eq $null)) {
		$Body.Add("tcpPortRange",$TcpPortRange)
	}

	if (!($ActivateNatTraversal -eq $null)) {
		$Body.Add("activateNatTraversal",$ActivateNatTraversal)
	}
	if ($ActivateNatTraversal) {
		$Body.Add("externalHost",$ExternalHost)

		if (!($NatDeviceHasFixedPortMapping -eq $null)) {
			$Body.Add("natDeviceHasFixedPortMapping",$NatDeviceHasFixedPortMapping)
		}
		if ($NatDeviceHasFixedPortMapping) {
			$Body.Add("fixedUdpTcpPort",$FixedUdpTcpPort)
			$Body.Add("fixedTlsPort",$FixedTlsPort)
			$Body.Add("fixedUdpPortRangeStart",$FixedUdpPortRangeStart)
			$Body.Add("fixedTcpPortRangeStart",$FixedTcpPortRangeStart)
		}
	}

	if (!($NodeViaSipTrunking -eq $null)) {
		$Body.Add("nodeViaSipTrunking",$NodeViaSipTrunking)
	}

	if (!($RemoteSipDomain -eq $null)) {
		$Body.Add("remoteSipDomain",$RemoteSipDomain)
	}

	if (!($LocalSipDomain -eq $null) -AND !($LocalSipDomain -eq '')) {
		$Body.Add("localSipDomain",$LocalSipDomain)
	}

	if (!($ActivateClientAuthentication -eq $null)) {
		$Body.Add("activateClientAuthentication",$ActivateClientAuthentication)
	}

	if (!($ClientCredentials -eq $null)) {
		$Body.Add("credentials",$ClientCredentials)
	}

	if (!($RegistrarUri -eq $null) -AND !($RegistrarUri -eq '')) {
		$Body.Add("registrarUri",$RegistrarUri)
	}

	if (!($AddressOfRecord -eq $null)) {
		$Body.Add("addressOfRecord",$AddressOfRecord)
	}

	if (!($ProxyUri -eq $null) -AND !($ProxyUri -eq '')) {
		$Body.Add("proxyUri",$ProxyUri)
	}

	if (!($PrimaryProxyUri -eq $null) -AND !($PrimaryProxyUri -eq '')) {
		$Body.Add("primaryProxyUri",$PrimaryProxyUri)
	}

	if (!($SecondaryProxyUri -eq $null) -AND !($SecondaryProxyUri -eq '')) {
		$Body.Add("secondaryProxyUri",$SecondaryProxyUri)
	}

	if (!($DefineAssertedUri -eq $null)) {
		$Body.Add("defineAssertedUri",$DefineAssertedUri)
	}
	if ($DefineAssertedUri) {
		$Body.Add("assertedUri",$AssertedUri)
	}

	if ($AssertedUriHost) {
		$Body.Add("assertedUriHost",$AssertedUriHost)
	}

	if (!($UseNetworkPeerWhitelist -eq $null)) {
		$Body.Add("useNetworkPeerWhitelist",$UseNetworkPeerWhitelist)
	}

	$Body.Add("includeRemoteSipDomainInWhitelist",$IncludeRemoteSipDomainInWhitelist)
	$Body.Add("includeSipRegistrarInWhitelist",$IncludeSipRegistrarInWhitelist)
	$Body.Add("includeProxyInWhitelist",$IncludeProxyInWhitelist)
	$Body.Add("networkPeerWhitelist",$NetworkPeerWhitelist)
	$Body.Add("allowOnlyNegotiatedPeersForRtpRtcp",$AllowOnlyNegotiatedPeersForRtpRtcp)

    if (!($IncomingDialStringRewriteRules -eq $null)) {
		if ($IncomingDialStringRewriteRules.GetType().Name -eq 'Object[]') {
			$Body.Add("incomingDialStringRewriteRules",$IncomingDialStringRewriteRules)
		} else {
			$Body.Add("incomingDialStringRewriteRules",@{0=$IncomingDialStringRewriteRules}.Values)
		}
	}

	if (!($OutgoingDialStringRewriteRules -eq $null)) {
		if ($OutgoingDialStringRewriteRules.GetType().Name -eq 'Object[]') {
			$Body.Add("outgoingDialStringRewriteRules",$OutgoingDialStringRewriteRules)
		} else {
			$Body.Add("outgoingDialStringRewriteRules",@{0=$OutgoingDialStringRewriteRules}.Values)
		}
	}

	if (!($RoutingDomains -eq $null)) {
		$Body.Add("routingDomains",$RoutingDomains)
	}

	if (!($DisplayName -eq $null) -AND !($DisplayName -eq '')) {
		$Body.Add("displayName",$DisplayName)
	}

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /nodes/addProvider -Body ($Body | ConvertTo-Json -Depth 99) -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get System Profiles
function Get-anynodeSystemNodeProfiles {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[switch] $CompleteData,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if ($CompleteData) {
		$Parameters.Add("completeData",$true)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/systemProfiles/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add System - deprecated parameters incomingManipulations and outgoingManipulations
function Add-anynodeSystemNode {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Profile,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $CreateNewNetworkController=$true,
		[Parameter()]
		[string] $NetworkControllerDisplayName,
		[Parameter()]
		[string] $NetworkControllerInterface='[Any Interface]',
		[Parameter()]
		[ValidateSet('[Any Address Version]','IN_ADDRESS_VERSION_4','IN_ADDRESS_VERSION_6')]
		[string] $NetworkControllerIpVersion='[Any Address Version]',
		[Parameter()]
		[string] $NetworkControllerIpAddress='[Any IP Address]',
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $EnableReverseDnsLookup,
		[Parameter()]
		[string] $NetworkControllerId,
		[Parameter()]
		[int] $UdpTcpPort,
		[Parameter()]
		[int] $TlsPort,
		[Parameter()]
		[psobject] $UdpPortRange,
		[Parameter()]
		[psobject] $TcpPortRange,
		[Parameter()]
		[ValidateSet("REGISTRAR","REGISTRATION","TRUNK")]
		[psobject] $SipInterconnection,
		[Parameter()]
		[psobject] $RemoteSipDomain,
		[Parameter()]
		[string] $LocalSipDomain,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $ActivateClientAuthentication,
		[Parameter()]
		[psobject] $ClientCredentials,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $ActivateServerAuthentication,
		[Parameter()]
		[psobject] $ServerCredentials,
		[Parameter()]
		[string] $RegistrarUri,
		[Parameter()]
		[string] $AddressOfRecord,
		[Parameter()]
		[psobject] $SipRegistrar,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $UseNetworkPeerWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludeRemoteSipDomainInWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludeSipRegistrarInWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludeOwnSubnetInWhitelist,
		[Parameter()]
		[psobject] $NetworkPeerWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $AllowOnlyNegotiatedPeersForRtpRtcp,
        [Parameter()]
		[psobject] $IncomingDialStringRewriteRules,
		[Parameter()]
		[psobject] $OutgoingDialStringRewriteRules,
		[Parameter()]
		[psobject] $RoutingDomains,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion=3,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password,
		[Parameter()]
		[psobject] $IncomingManipulations,
		[Parameter()]
		[psobject] $OutgoingManipulations
	)

	$Parameters=@{profile=$Profile}

	$Body=@{}

	$Body.Add("createNewNetworkController",$CreateNewNetworkController)
	if ($CreateNewNetworkController) {
		if (!($NetworkControllerDisplayName -eq $null) -AND !($NetworkControllerDisplayName -eq '')) {
			$Body.Add("networkControllerDisplayName",$NetworkControllerDisplayName)
		}
		$Body.Add("networkControllerInterface",$NetworkControllerInterface)
		$Body.Add("networkControllerIpVersion",$NetworkControllerIpVersion)
		$Body.Add("networkControllerIpAddress",$NetworkControllerIpAddress)
		if (!($EnableReverseDnsLookup -eq $null)) {
			$Body.Add("enableReverseDnsLookup",$EnableReverseDnsLookup)
		}
	}

	if (!($NetworkControllerId -eq $null) -AND !($NetworkControllerId -eq '')) {
		$Body.Add("networkControllerId",$NetworkControllerId)
	}

	if (!($UdpTcpPort -eq $null) -AND $UdpTcpPort -gt 0) {
		$Body.Add("udpTcpPort",$UdpTcpPort)
	}

	if (!($TlsPort -eq $null) -AND $TlsPort -gt 0) {
		$Body.Add("tlsPort",$TlsPort)
	}

	if (!($UdpPortRange -eq $null)) {
		$Body.Add("udpPortRange",$UdpPortRange)
	}

	if (!($TcpPortRange -eq $null)) {
		$Body.Add("tcpPortRange",$TcpPortRange)
	}

	if (!($SipInterconnection -eq $null)) {
		$Body.Add("sipInterconnection",$SipInterconnection)
	}

	if (!($RemoteSipDomain -eq $null)) {
		$Body.Add("remoteSipDomain",$RemoteSipDomain)
	}

	if (!($LocalSipDomain -eq $null) -AND !($LocalSipDomain -eq '')) {
		$Body.Add("localSipDomain",$LocalSipDomain)
	}

	if (!($ActivateClientAuthentication -eq $null)) {
		$Body.Add("activateClientAuthentication",$ActivateClientAuthentication)
	}
	if (!($ClientCredentials -eq $null)) {
		$Body.Add("clientCredentials",$ClientCredentials)
	}

	if (!($ActivateServerAuthentication -eq $null)) {
		$Body.Add("activateServerAuthentication",$ActivateServerAuthentication)
	}
	if (!($ServerCredentials -eq $null)) {
		$Body.Add("serverCredentials",$ServerCredentials)
	}

	if (!($RegistrarUri -eq $null) -AND !($RegistrarUri -eq '')) {
		$Body.Add("registrarUri",$RegistrarUri)
	}

	if (!($AddressOfRecord -eq $null)) {
		$Body.Add("addressOfRecord",$AddressOfRecord)
	}

	if (!($SipRegistrar -eq $null)) {
		$Body.Add("sipRegistrar",$SipRegistrar)
	}

	if (!($UseNetworkPeerWhitelist -eq $null)) {
		$Body.Add("useNetworkPeerWhitelist",$UseNetworkPeerWhitelist)
	}

	$Body.Add("includeRemoteSipDomainInWhitelist",$IncludeRemoteSipDomainInWhitelist)
	$Body.Add("includeSipRegistrarInWhitelist",$IncludeSipRegistrarInWhitelist)
	$Body.Add("includeProxyInWhitelist",$IncludeProxyInWhitelist)
	$Body.Add("includeOwnSubnetInWhitelist",$IncludeOwnSubnetInWhitelist)
	$Body.Add("networkPeerWhitelist",$NetworkPeerWhitelist)
	$Body.Add("allowOnlyNegotiatedPeersForRtpRtcp",$AllowOnlyNegotiatedPeersForRtpRtcp)

	if (!($IncomingManipulations -eq $null)) {
		if ($IncomingManipulations.GetType().Name -eq 'Object[]') {
			$Body.Add("incomingManipulations",$IncomingManipulations)
		} else {
			$Body.Add("incomingManipulations",@{0=$IncomingManipulations}.Values)
		}
	}

	if (!($OutgoingManipulations -eq $null)) {
		if ($OutgoingManipulations.GetType().Name -eq 'Object[]') {
			$Body.Add("outgoingManipulations",$OutgoingManipulations)
		} else {
			$Body.Add("outgoingManipulations",@{0=$OutgoingManipulations}.Values)
		}
	}

	if (!($IncomingDialStringRewriteRules -eq $null)) {
		if ($IncomingDialStringRewriteRules.GetType().Name -eq 'Object[]') {
			$Body.Add("incomingDialStringRewriteRules",$IncomingDialStringRewriteRules)
		} else {
			$Body.Add("incomingDialStringRewriteRules",@{0=$IncomingDialStringRewriteRules}.Values)
		}
	}

	if (!($OutgoingDialStringRewriteRules -eq $null)) {
		if ($OutgoingDialStringRewriteRules.GetType().Name -eq 'Object[]') {
			$Body.Add("outgoingDialStringRewriteRules",$OutgoingDialStringRewriteRules)
		} else {
			$Body.Add("outgoingDialStringRewriteRules",@{0=$OutgoingDialStringRewriteRules}.Values)
		}
	}

	if (!($RoutingDomains -eq $null)) {
		$Body.Add("routingDomains",$RoutingDomains)
	}

	if (!($DisplayName -eq $null) -AND !($DisplayName -eq '')) {
		$Body.Add("displayName",$DisplayName)
	}

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /nodes/addSystem -Body ($Body | ConvertTo-Json -Depth 99) -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	DEPRECATED
function Add-anynodeSystemNodeV2 {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Profile,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $CreateNewNetworkController=$true,
		[Parameter()]
		[string] $NetworkControllerDisplayName,
		[Parameter()]
		[string] $NetworkControllerInterface='[Any Interface]',
		[Parameter()]
		[ValidateSet('[Any Address Version]','IN_ADDRESS_VERSION_4','IN_ADDRESS_VERSION_6')]
		[string] $NetworkControllerIpVersion='[Any Address Version]',
		[Parameter()]
		[string] $NetworkControllerIpAddress='[Any IP Address]',
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $EnableReverseDnsLookup,
		[Parameter()]
		[string] $NetworkControllerId,
		[Parameter()]
		[int] $UdpTcpPort,
		[Parameter()]
		[int] $TlsPort,
		[Parameter()]
		[psobject] $UdpPortRange,
		[Parameter()]
		[psobject] $TcpPortRange,
		[Parameter()]
		[ValidateSet("REGISTRAR","REGISTRATION","TRUNK")]
		[psobject] $SipInterconnection,
		[Parameter()]
		[psobject] $RemoteSipDomain,
		[Parameter()]
		[string] $LocalSipDomain,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $ActivateClientAuthentication,
		[Parameter()]
		[psobject] $ClientCredentials,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $ActivateServerAuthentication,
		[Parameter()]
		[psobject] $ServerCredentials,
		[Parameter()]
		[string] $RegistrarUri,
		[Parameter()]
		[string] $AddressOfRecord,
		[Parameter()]
		[psobject] $SipRegistrar,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $UseNetworkPeerWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludeRemoteSipDomainInWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludeSipRegistrarInWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludeOwnSubnetInWhitelist,
		[Parameter()]
		[psobject] $NetworkPeerWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $AllowOnlyNegotiatedPeersForRtpRtcp,
        [Parameter()]
		[psobject] $IncomingDialStringRewriteRules,
		[Parameter()]
		[psobject] $OutgoingDialStringRewriteRules,
		[Parameter()]
		[psobject] $RoutingDomains,
		[Parameter()]
		[string] $DisplayName,
        [Parameter()]
        [string] $GooglePbxTrunkSecretKey,
        [Parameter()]
		[ValidateSet("NA","EMEA","AUS","APAC","LATAM","CHINA","JAPAN")]
        [psobject] $Region,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion=3,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{profile=$Profile}

	$Body=@{}

	$Body.Add("createNewNetworkController",$CreateNewNetworkController)
	if ($CreateNewNetworkController) {
		if (!($NetworkControllerDisplayName -eq $null) -AND !($NetworkControllerDisplayName -eq '')) {
			$Body.Add("networkControllerDisplayName",$NetworkControllerDisplayName)
		}
		$Body.Add("networkControllerInterface",$NetworkControllerInterface)
		$Body.Add("networkControllerIpVersion",$NetworkControllerIpVersion)
		$Body.Add("networkControllerIpAddress",$NetworkControllerIpAddress)
		if (!($EnableReverseDnsLookup -eq $null)) {
			$Body.Add("enableReverseDnsLookup",$EnableReverseDnsLookup)
		}
	}

	if (!($NetworkControllerId -eq $null) -AND !($NetworkControllerId -eq '')) {
		$Body.Add("networkControllerId",$NetworkControllerId)
	}

	if (!($UdpTcpPort -eq $null) -AND $UdpTcpPort -gt 0) {
		$Body.Add("udpTcpPort",$UdpTcpPort)
	}

	if (!($TlsPort -eq $null) -AND $TlsPort -gt 0) {
		$Body.Add("tlsPort",$TlsPort)
	}

	if (!($UdpPortRange -eq $null)) {
		$Body.Add("udpPortRange",$UdpPortRange)
	}

	if (!($TcpPortRange -eq $null)) {
		$Body.Add("tcpPortRange",$TcpPortRange)
	}

	if (!($SipInterconnection -eq $null)) {
		$Body.Add("sipInterconnection",$SipInterconnection)
	}

	if (!($RemoteSipDomain -eq $null)) {
		$Body.Add("remoteSipDomain",$RemoteSipDomain)
	}

	if (!($LocalSipDomain -eq $null) -AND !($LocalSipDomain -eq '')) {
		$Body.Add("localSipDomain",$LocalSipDomain)
	}

	if (!($ActivateClientAuthentication -eq $null)) {
		$Body.Add("activateClientAuthentication",$ActivateClientAuthentication)
	}
	if (!($ClientCredentials -eq $null)) {
		$Body.Add("clientCredentials",$ClientCredentials)
	}

	if (!($ActivateServerAuthentication -eq $null)) {
		$Body.Add("activateServerAuthentication",$ActivateServerAuthentication)
	}
	if (!($ServerCredentials -eq $null)) {
		$Body.Add("serverCredentials",$ServerCredentials)
	}

	if (!($RegistrarUri -eq $null) -AND !($RegistrarUri -eq '')) {
		$Body.Add("registrarUri",$RegistrarUri)
	}

	if (!($AddressOfRecord -eq $null)) {
		$Body.Add("addressOfRecord",$AddressOfRecord)
	}

	if (!($SipRegistrar -eq $null)) {
		$Body.Add("sipRegistrar",$SipRegistrar)
	}

	if (!($UseNetworkPeerWhitelist -eq $null)) {
		$Body.Add("useNetworkPeerWhitelist",$UseNetworkPeerWhitelist)
	}

	$Body.Add("includeRemoteSipDomainInWhitelist",$IncludeRemoteSipDomainInWhitelist)
	$Body.Add("includeSipRegistrarInWhitelist",$IncludeSipRegistrarInWhitelist)
	$Body.Add("includeProxyInWhitelist",$IncludeProxyInWhitelist)
	$Body.Add("includeOwnSubnetInWhitelist",$IncludeOwnSubnetInWhitelist)
	$Body.Add("networkPeerWhitelist",$NetworkPeerWhitelist)
	$Body.Add("allowOnlyNegotiatedPeersForRtpRtcp",$AllowOnlyNegotiatedPeersForRtpRtcp)

    if (!($IncomingDialStringRewriteRules -eq $null)) {
		if ($IncomingDialStringRewriteRules.GetType().Name -eq 'Object[]') {
			$Body.Add("incomingDialStringRewriteRules",$IncomingDialStringRewriteRules)
		} else {
			$Body.Add("incomingDialStringRewriteRules",@{0=$IncomingDialStringRewriteRules}.Values)
		}
	}

	if (!($OutgoingDialStringRewriteRules -eq $null)) {
		if ($OutgoingDialStringRewriteRules.GetType().Name -eq 'Object[]') {
			$Body.Add("outgoingDialStringRewriteRules",$OutgoingDialStringRewriteRules)
		} else {
			$Body.Add("outgoingDialStringRewriteRules",@{0=$OutgoingDialStringRewriteRules}.Values)
		}
	}

    if(!($GooglePbxTrunkSecretKey -eq $null)) {
        $Body.Add("googlePbcTrunkSecretKey", $GooglePbxTrunkSecretKey)
    }

    if(!($Region -eq $null)) {
        $Body.Add("region", $Region)
    }

	if (!($RoutingDomains -eq $null)) {
		$Body.Add("routingDomains",$RoutingDomains)
	}

	if (!($DisplayName -eq $null) -AND !($DisplayName -eq '')) {
		$Body.Add("displayName",$DisplayName)
	}

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /nodes/addSystem -Body ($Body | ConvertTo-Json -Depth 99) -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add MS Teams Tenant - deprecated parameters incomingManipulations and outgoingManipulations
function Add-anynodeMsTeamsTenantNode {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $CreateNewTeamsCarrierNode=$false,
		[Parameter()]
		[string] $TeamsCarrierNodeId,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $CreateNewNetworkController=$true,
		[Parameter()]
		[string] $NetworkControllerDisplayName,
		[Parameter()]
		[string] $NetworkControllerInterface='[Any Interface]',
		[Parameter()]
		[ValidateSet('[Any Address Version]','IN_ADDRESS_VERSION_4','IN_ADDRESS_VERSION_6')]
		[string] $NetworkControllerIpVersion='[Any Address Version]',
		[Parameter()]
		[string] $NetworkControllerIpAddress='[Any IP Address]',
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $EnableReverseDnsLookup,
		[Parameter()]
		[string] $NetworkControllerId,
		[Parameter()]
		[int] $UdpTcpPort,
		[Parameter()]
		[int] $TlsPort,
		[Parameter()]
		[psobject] $UdpPortRange,
		[Parameter()]
		[psobject] $TcpPortRange,
		[Parameter()]
		[string] $SbcFqdn,
		[Parameter()]
		[string] $TenantFqdn,
        [Parameter()]
		[psobject] $IncomingDialStringRewriteRules,
		[Parameter()]
		[psobject] $OutgoingDialStringRewriteRules,
		[Parameter()]
		[psobject] $RoutingDomains,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion=3,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password,
		[Parameter()]
		[psobject] $IncomingManipulations,
		[Parameter()]
		[psobject] $OutgoingManipulations
	)

    $Profile=Get-anynodeSystemNodeProfiles -Address $Address -Port $Port -Credential $Credential -Username $Username -Password $Password | Where-Object -Property "displayName" -EQ "Microsoft Teams Direct Routing Carrier"

	$Parameters=@{profile=$Profile.id}

	$Body=@{}

	$Body.Add("createNewTeamsCarrierNode",$CreateNewTeamsCarrierNode)
	if ($CreateNewTeamsCarrierNode) {

		$Body.Add("createNewNetworkController",$CreateNewNetworkController)
		if ($CreateNewNetworkController) {
			if (!($NetworkControllerDisplayName -eq $null) -AND !($NetworkControllerDisplayName -eq '')) {
				$Body.Add("networkControllerDisplayName",$NetworkControllerDisplayName)
			}
			$Body.Add("networkControllerInterface",$NetworkControllerInterface)
			$Body.Add("networkControllerIpVersion",$NetworkControllerIpVersion)
			$Body.Add("networkControllerIpAddress",$NetworkControllerIpAddress)
			if (!($EnableReverseDnsLookup -eq $null)) {
				$Body.Add("enableReverseDnsLookup",$EnableReverseDnsLookup)
			}
		}

		if (!($NetworkControllerId -eq $null) -AND !($NetworkControllerId -eq '')) {
			$Body.Add("networkControllerId",$NetworkControllerId)
		}
	
		if (!($UdpTcpPort -eq $null) -AND $UdpTcpPort -gt 0) {
			$Body.Add("udpTcpPort",$UdpTcpPort)
		}
	
		if (!($TlsPort -eq $null) -AND $TlsPort -gt 0) {
			$Body.Add("tlsPort",$TlsPort)
		}
	
		if (!($UdpPortRange -eq $null)) {
			$Body.Add("udpPortRange",$UdpPortRange)
		}
	
		if (!($TcpPortRange -eq $null)) {
			$Body.Add("tcpPortRange",$TcpPortRange)
		}

		if (!($SbcFqdn -eq $null) -AND !($SbcFqdn -eq '')) {
			$Body.Add("sbcFqdn",$SbcFqdn)
		}
	}

	if (!($TeamsCarrierNodeId -eq $null) -AND !($TeamsCarrierNodeId -eq '')) {
		$Body.Add("teamsCarrierNodeId",$TeamsCarrierNodeId)
	}

	if (!($TenantFqdn -eq $null) -AND !($TenantFqdn -eq '')) {
		$Body.Add("tenantFqdn",$TenantFqdn)
	}

	if (!($IncomingManipulations -eq $null)) {
		if ($IncomingManipulations.GetType().Name -eq 'Object[]') {
			$Body.Add("incomingManipulations",$IncomingManipulations)
		} else {
			$Body.Add("incomingManipulations",@{0=$IncomingManipulations}.Values)
		}
	}

	if (!($OutgoingManipulations -eq $null)) {
		if ($OutgoingManipulations.GetType().Name -eq 'Object[]') {
			$Body.Add("outgoingManipulations",$OutgoingManipulations)
		} else {
			$Body.Add("outgoingManipulations",@{0=$OutgoingManipulations}.Values)
		}
	}

	if (!($IncomingDialStringRewriteRules -eq $null)) {
		if ($IncomingDialStringRewriteRules.GetType().Name -eq 'Object[]') {
			$Body.Add("incomingDialStringRewriteRules",$IncomingDialStringRewriteRules)
		} else {
			$Body.Add("incomingDialStringRewriteRules",@{0=$IncomingDialStringRewriteRules}.Values)
		}
	}

	if (!($OutgoingDialStringRewriteRules -eq $null)) {
		if ($OutgoingDialStringRewriteRules.GetType().Name -eq 'Object[]') {
			$Body.Add("outgoingDialStringRewriteRules",$OutgoingDialStringRewriteRules)
		} else {
			$Body.Add("outgoingDialStringRewriteRules",@{0=$OutgoingDialStringRewriteRules}.Values)
		}
	}

	if (!($RoutingDomains -eq $null)) {
		$Body.Add("routingDomains",$RoutingDomains)
	}

	if (!($DisplayName -eq $null) -AND !($DisplayName -eq '')) {
		$Body.Add("displayName",$DisplayName)
	}

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /nodes/addSystem -Body ($Body | ConvertTo-Json -Depth 99) -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	DEPRECATED!
function Add-anynodeMsTeamsTenantNodeV2 {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $CreateNewTeamsCarrierNode=$false,
		[Parameter()]
		[string] $TeamsCarrierNodeId,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $CreateNewNetworkController=$true,
		[Parameter()]
		[string] $NetworkControllerDisplayName,
		[Parameter()]
		[string] $NetworkControllerInterface='[Any Interface]',
		[Parameter()]
		[ValidateSet('[Any Address Version]','IN_ADDRESS_VERSION_4','IN_ADDRESS_VERSION_6')]
		[string] $NetworkControllerIpVersion='[Any Address Version]',
		[Parameter()]
		[string] $NetworkControllerIpAddress='[Any IP Address]',
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $EnableReverseDnsLookup,
		[Parameter()]
		[string] $NetworkControllerId,
		[Parameter()]
		[int] $UdpTcpPort,
		[Parameter()]
		[int] $TlsPort,
		[Parameter()]
		[psobject] $UdpPortRange,
		[Parameter()]
		[psobject] $TcpPortRange,
		[Parameter()]
		[string] $SbcFqdn,
		[Parameter()]
		[string] $TenantFqdn,
        [Parameter()]
		[psobject] $IncomingDialStringRewriteRules,
		[Parameter()]
		[psobject] $OutgoingDialStringRewriteRules,
		[Parameter()]
		[psobject] $RoutingDomains,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion=3,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

    $Profile=Get-anynodeSystemNodeProfiles -Address $Address -Port $Port -Credential $Credential -Username $Username -Password $Password | Where-Object -Property "displayName" -EQ "Microsoft Teams Direct Routing Carrier"

	$Parameters=@{profile=$Profile.id}

	$Body=@{}

	$Body.Add("createNewTeamsCarrierNode",$CreateNewTeamsCarrierNode)
	if ($CreateNewTeamsCarrierNode) {

		$Body.Add("createNewNetworkController",$CreateNewNetworkController)
		if ($CreateNewNetworkController) {
			if (!($NetworkControllerDisplayName -eq $null) -AND !($NetworkControllerDisplayName -eq '')) {
				$Body.Add("networkControllerDisplayName",$NetworkControllerDisplayName)
			}
			$Body.Add("networkControllerInterface",$NetworkControllerInterface)
			$Body.Add("networkControllerIpVersion",$NetworkControllerIpVersion)
			$Body.Add("networkControllerIpAddress",$NetworkControllerIpAddress)
			if (!($EnableReverseDnsLookup -eq $null)) {
				$Body.Add("enableReverseDnsLookup",$EnableReverseDnsLookup)
			}
		}

		if (!($NetworkControllerId -eq $null) -AND !($NetworkControllerId -eq '')) {
			$Body.Add("networkControllerId",$NetworkControllerId)
		}
	
		if (!($UdpTcpPort -eq $null) -AND $UdpTcpPort -gt 0) {
			$Body.Add("udpTcpPort",$UdpTcpPort)
		}
	
		if (!($TlsPort -eq $null) -AND $TlsPort -gt 0) {
			$Body.Add("tlsPort",$TlsPort)
		}
	
		if (!($UdpPortRange -eq $null)) {
			$Body.Add("udpPortRange",$UdpPortRange)
		}
	
		if (!($TcpPortRange -eq $null)) {
			$Body.Add("tcpPortRange",$TcpPortRange)
		}

		if (!($SbcFqdn -eq $null) -AND !($SbcFqdn -eq '')) {
			$Body.Add("sbcFqdn",$SbcFqdn)
		}
	}

	if (!($TeamsCarrierNodeId -eq $null) -AND !($TeamsCarrierNodeId -eq '')) {
		$Body.Add("teamsCarrierNodeId",$TeamsCarrierNodeId)
	}

	if (!($TenantFqdn -eq $null) -AND !($TenantFqdn -eq '')) {
		$Body.Add("tenantFqdn",$TenantFqdn)
	}

    if (!($IncomingDialStringRewriteRules -eq $null)) {
		if ($IncomingDialStringRewriteRules.GetType().Name -eq 'Object[]') {
			$Body.Add("incomingDialStringRewriteRules",$IncomingDialStringRewriteRules)
		} else {
			$Body.Add("incomingDialStringRewriteRules",@{0=$IncomingDialStringRewriteRules}.Values)
		}
	}

	if (!($OutgoingDialStringRewriteRules -eq $null)) {
		if ($OutgoingDialStringRewriteRules.GetType().Name -eq 'Object[]') {
			$Body.Add("outgoingDialStringRewriteRules",$OutgoingDialStringRewriteRules)
		} else {
			$Body.Add("outgoingDialStringRewriteRules",@{0=$OutgoingDialStringRewriteRules}.Values)
		}
	}

	if (!($RoutingDomains -eq $null)) {
		$Body.Add("routingDomains",$RoutingDomains)
	}

	if (!($DisplayName -eq $null) -AND !($DisplayName -eq '')) {
		$Body.Add("displayName",$DisplayName)
	}

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /nodes/addSystem -Body ($Body | ConvertTo-Json -Depth 99) -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get Incoming Manipulations
function Get-anynodeNodeIncomingManipulations {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[switch] $IncludeDefaults,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}
	if ($IncludeDefaults) {
		$Parameters.Add("includeDefaults",$true)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/incomingManipulations/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set Incoming Manipulations
function Set-anynodeNodeIncomingManipulations {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[psobject] $Manipulations,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	$Body=($Manipulations | ConvertTo-Json)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/incomingManipulations/set -Parameters $Parameters -Body $Body -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add Incoming Manipulations
function Add-anynodeNodeIncomingManipulations {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[psobject] $Manipulations,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	$Body=($Manipulations | ConvertTo-Json)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/incomingManipulations/add -Parameters $Parameters -Body $Body -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Remove Incoming Manipulation
function Remove-anynodeNodeIncomingManipulation {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[int] $Index=-1,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	if ($Index -gt -1) {
		$Parameters.Add("index",$Index)
	} elseif ($DisplayName -ne $null -AND $DisplayName -ne '') {
		$Parameters.Add("displayName",$DisplayName)
	}

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /nodes/incomingManipulations/remove -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get Outgoing Manipulations
function Get-anynodeNodeOutgoingManipulations {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[switch] $IncludeDefaults,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}
	if ($IncludeDefaults) {
		$Parameters.Add("includeDefaults",$true)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/outgoingManipulations/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set Outgoing Manipulations
function Set-anynodeNodeOutgoingManipulations {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[psobject] $Manipulations,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	$Body=($Manipulations | ConvertTo-Json)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/outgoingManipulations/set -Parameters $Parameters -Body $Body -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add Outgoing Manipulations
function Add-anynodeNodeOutgoingManipulations {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[psobject] $Manipulations,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	$Body=($Manipulations | ConvertTo-Json)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/outgoingManipulations/add -Parameters $Parameters -Body $Body -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Remove Outgoing Manipulation
function Remove-anynodeNodeOutgoingManipulation {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[int] $Index=-1,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	if ($Index -gt -1) {
		$Parameters.Add("index",$Index)
	} elseif ($DisplayName -ne $null -AND $DisplayName -ne '') {
		$Parameters.Add("displayName",$DisplayName)
	}

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /nodes/outgoingManipulations/remove -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get Incoming DialStringRewriteRules
function Get-anynodeNodeIncomingDialStringRewriteRules {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[switch] $IncludeDefaults,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}
	if ($IncludeDefaults) {
		$Parameters.Add("includeDefaults",$true)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/incomingDialStringRewriteRules/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set Incoming DialStringRewriteRules
function Set-anynodeNodeIncomingDialStringRewriteRules {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[psobject] $DialStringRewriteRules,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	$Body= (ConvertTo-Json -InputObject @($DialStringRewriteRules))

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/incomingDialStringRewriteRules/set -Parameters $Parameters -Body $Body -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add Incoming DialStringRewriteRules
function Add-anynodeNodeIncomingDialStringRewriteRules {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[psobject] $DialStringRewriteRules,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	$Body= (ConvertTo-Json -InputObject @($DialStringRewriteRules))

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/incomingDialStringRewriteRules/add -Parameters $Parameters -Body $Body -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Remove Incoming DialStringRewriteRules
function Remove-anynodeNodeIncomingDialStringRewriteRules {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[int] $Index=-1,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	if ($Index -gt -1) {
		$Parameters.Add("index",$Index)
	} elseif ($DisplayName -ne $null -AND $DisplayName -ne '') {
		$Parameters.Add("displayName",$DisplayName)
	}

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /nodes/incomingDialStringRewriteRules/remove -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get Outgoing DialStringRewriteRules
function Get-anynodeNodeOutgoingDialStringRewriteRules {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[switch] $IncludeDefaults,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}
	if ($IncludeDefaults) {
		$Parameters.Add("includeDefaults",$true)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/outgoingDialStringRewriteRules/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set Outgoing DialStringRewriteRules
function Set-anynodeNodeOutgoingDialStringRewriteRules {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[psobject] $DialStringRewriteRules,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	$Body= (ConvertTo-Json -InputObject @($DialStringRewriteRules))

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/outgoingDialStringRewriteRules/set -Parameters $Parameters -Body $Body -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
	
}

#	Add Outgoing DialStringRewriteRules
function Add-anynodeNodeOutgoingDialStringRewriteRules {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[psobject] $DialStringRewriteRules,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	$Body= (ConvertTo-Json -InputObject @($DialStringRewriteRules))

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/outgoingDialStringRewriteRules/add -Parameters $Parameters -Body $Body -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Remove Outgoing DialStringRewriteRules
function Remove-anynodeNodeOutgoingDialStringRewriteRules {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[int] $Index=-1,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	if ($Index -gt -1) {
		$Parameters.Add("index",$Index)
	} elseif ($DisplayName -ne $null -AND $DisplayName -ne '') {
		$Parameters.Add("displayName",$DisplayName)
	}

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /nodes/outgoingDialStringRewriteRules/remove -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get Network Peer Whitelist
function Get-anynodeNodeNetworkPeerWhitelist {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/networkPeerWhitelist/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set Network Peer Whitelist
function Set-anynodeNodeNetworkPeerWhitelist {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter(mandatory=$true)]
		[psobject] $NetworkPeerWhitelist,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	$Body=($NetworkPeerWhitelist | ConvertTo-Json)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/networkPeerWhitelist/set -Parameters $Parameters -Body $Body -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add Network Peer Whitelist Entry
function Add-anynodeNodeNetworkPeerWhitelistEntry {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter(mandatory=$true)]
		[psobject] $NetworkPeerWhitelist,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	$Body=($NetworkPeerWhitelist | ConvertTo-Json)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/networkPeerWhitelist/add -Parameters $Parameters -Body $Body -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Remove Network Peer Whitelist Entry
function Remove-anynodeNodeNetworkPeerWhitelistEntry {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[int] $Index=-1,
		[Parameter()]
		[string] $Hostname,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	if ($Index -gt -1) {
		$Parameters.Add("index",$Index)
	} elseif ($Hostname -ne $null -AND $Hostname -ne '') {
		$Parameters.Add("hostname",$Hostname)
	}

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /nodes/networkPeerWhitelist/remove -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Remove
function Remove-anynodeNode {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[switch] $Force,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}
	if ($Force) {
		$Parameters.Add("force",$true)
	}

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /nodes/remove -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get MS Teams Connectivities
function Get-anynodeMsTeamsCarriers {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/msTeamsCarriers/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Node Certificates
#	Get
function Get-anynodeNodeCertificate {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[ValidateSet('ISO-8601','XZWO')]
		[string] $DateFormat,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}
	if ($DateFormat -ne $null -AND $DateFormat -ne '') {
		$Parameters.Add('dateFormat',$DateFormat)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/certificates/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set
function Set-anynodeNodeCertificate {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[string] $CertificateFilePath,
		[Parameter()]
		[string] $KeyFilePath,
		[Parameter()]
		[string] $CertificatePassword,
		[Parameter()]
		[ValidateSet("CERTIFICATES_AND_PRIVATE_KEY","CERTIFICATES_ONLY","PRIVATE_KEY_ONLY")]
		[string] $Mode='CERTIFICATES_AND_PRIVATE_KEY',
		[Parameter()]
		[ValidateSet('ISO-8601','XZWO')]
		[string] $DateFormat,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}
	if (!($CertificatePassword -eq '')) {
		$Parameters.Add("password",$CertificatePassword)
	}

	if ($DateFormat -ne $null -AND $DateFormat -ne '') {
		$Parameters.Add('dateFormat',$DateFormat)
	}
	if (!($CertificateFilePath -eq '')) {
		$RequestParams=$Parameters.Clone()
		if ($Mode -eq 'CERTIFICATES_AND_PRIVATE_KEY') {
			$RequestParams.Add("mode",'CERTIFICATES_ONLY')
		} elseif (!($Mode -eq '')) {
			$RequestParams.Add("mode",$Mode)
		}

		Invoke-anynodeUpload -Address $Address -Port $Port -URITail /nodes/certificates/set -Parameters $RequestParams -Filenames @($CertificateFilePath) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
	}

	if (!($KeyFilePath -eq '')) {
		$RequestParams=$Parameters.Clone()
		if ($Mode -eq 'CERTIFICATES_AND_PRIVATE_KEY') {
			$RequestParams.Add("mode",'PRIVATE_KEY_ONLY')
		} elseif (!($Mode -eq '')) {
			$RequestParams.Add("mode",$Mode)
		}

		Invoke-anynodeUpload -Address $Address -Port $Port -URITail /nodes/certificates/set -Parameters $RequestParams -Filenames @($KeyFilePath) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
	}

# This doesn't work because anynode only accepts one file at a time.
#	$Files=@{}
#	if (!($CertificateFilePath -eq '')) {
#		$Files.Add(1,$CertificateFilePath)
#	}
#
#	if (!($KeyFilePath -eq '')) {
#		$Files.Add(2,$KeyFilePath)
#	}
#
#	if ($Files.Count -eq 0) {
#		echo 'Path(s) to Certificate and/or Private Key must be given for this call to work.'
#		return
#	}
#
#	Invoke-anynodeUpload -Address $Address -Port $Port -URITail /nodes/certificates/set -Parameters $Parameters -Filenames $Files.Values -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Node Chain Certificates
#	Get
function Get-anynodeNodeChainCertificates {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[ValidateSet('ISO-8601','XZWO')]
		[string] $DateFormat,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}
	if ($DateFormat -ne $null -AND $DateFormat -ne '') {
		$Parameters.Add('dateFormat',$DateFormat)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/certificateChain/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set
function Set-anynodeNodeChainCertificates {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter(mandatory=$true)]
		[string[]] $CertificateFiles,
		[Parameter()]
		[ValidateSet('ISO-8601','XZWO')]
		[string] $DateFormat,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}
	if ($DateFormat -ne $null -AND $DateFormat -ne '') {
		$Parameters.Add('dateFormat',$DateFormat)
	}

	$Body=''

	foreach ($File in $CertificateFiles) {
		if ($File -eq $null) {
			continue
		}

		try {
			$Body+=Get-Content $File
		} catch {
			echo $_.Exception
		}
	}

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/certificateChain/set -Parameters $Parameters -Body $Body -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add
function Add-anynodeNodeChainCertificate {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter(mandatory=$true)]
		[string] $CertificateFilePath,
		[Parameter()]
		[ValidateSet('ISO-8601','XZWO')]
		[string] $DateFormat,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}
	if ($DateFormat -ne $null -AND $DateFormat -ne '') {
		$Parameters.Add('dateFormat',$DateFormat)
	}

	Invoke-anynodeUpload -Address $Address -Port $Port -URITail /nodes/certificateChain/add -Filenames @($CertificateFilePath) -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Remove
function Remove-anynodeNodeChainCertificate {
	[CmdletBinding(DefaultParameterSetName='Index')]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter(mandatory=$true, ParameterSetName='Index')]
		[int] $Index,
		[Parameter(mandatory=$true, ParameterSetName='Host')]
		[string] $Host,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	if ($Host -ne $null -AND $Host -ne '') {
		$Parameters.Add('host',$Host)
	} else {
		$Parameters.Add('index',$Index)
	}

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /nodes/certificateChain/remove -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Clear
function Clear-anynodeNodeChainCertificates {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /nodes/certificateChain/clear -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Request
function Request-anynodeNodeChainCertificates {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[ValidateSet('ISO-8601','XZWO')]
		[string] $DateFormat,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}
	if ($DateFormat -ne $null -AND $DateFormat -ne '') {
		$Parameters.Add('dateFormat',$DateFormat)
	}

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /nodes/certificateChain/request -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Node Trusted Certificates
#	Get
function Get-anynodeNodeTrustedCertificates {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[ValidateSet('ISO-8601','XZWO')]
		[string] $DateFormat,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}
	if ($DateFormat -ne $null -AND $DateFormat -ne '') {
		$Parameters.Add('dateFormat',$DateFormat)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/trustedCertificates/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add
function Add-anynodeNodeTrustedCertificate {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter(mandatory=$true)]
		[string] $CertificateFilePath,
		[Parameter()]
		[string] $CertificatePassword,
		[Parameter()]
		[ValidateSet('ISO-8601','XZWO')]
		[string] $DateFormat,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}
	if ($DateFormat -ne $null -AND $DateFormat -ne '') {
		$Parameters.Add('dateFormat',$DateFormat)
	}

	if (!($CertificatePassword -eq '')) {
		$Parameters.Add("password",$CertificatePassword)
	}

	$Files=@{}
	if (!($CertificateFilePath -eq '')) {
		$Files.Add(1,$CertificateFilePath)
	}

	if ($Files.Count -eq 0) {
		echo 'Path to a Certificate must be given for this call to work.'
		return
	}

	Invoke-anynodeUpload -Address $Address -Port $Port -URITail /nodes/trustedCertificates/add -Parameters $Parameters -Filenames $Files.Values -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Clear
function Clear-anynodeNodeTrustedCertificates {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /nodes/trustedCertificates/clear -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Node FQDN
#	Get
function Get-anynodeNodeFQDN {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/tenants/fqdn/get -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set
function Set-anynodeNodeFQDN {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter(mandatory=$true)]
		[string] $NewFQDN,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/tenants/fqdn/set -Body $NewFQDN -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Node override hostname
#	Get
function Get-anynodeNodeOverrideHostname {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/overrideHostname/get -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set
function Set-anynodeNodeOverrideHostname {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter(mandatory=$true)]
		[string] $NewOverrideHostname,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/overrideHostname/set -Body $NewOverrideHostname -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Node Port
#	Get
function Get-anynodeNodePort {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/port/get -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set
function Set-anynodeNodePort {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter(mandatory=$true)]
		[int] $NewPort,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/port/set -Body $NewPort -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Node TLS Port
#	Get
function Get-anynodeNodeTlsPort {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/tlsPort/get -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set
function Set-anynodeNodeTlsPort {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter(mandatory=$true)]
		[int] $NewPort,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/tlsPort/set -Body $NewPort -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Node Transport Connection
#	Get
function Get-anynodeNodeTransportConnections {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/transportConnections/get -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add
function Add-anynodeNodeTransportConnection {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter(mandatory=$true)]
		[psobject]$MatchingCondition,
		[Parameter(mandatory=$true)]
		[psobject]$TransportConnectionId,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}
	$Body.Add("transportConnectionId",$TransportConnectionId)
	$Body.Add("matchingCondition",$MatchingCondition)

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /nodes/transportConnections/add -Body ($Body | ConvertTo-Json -Depth 99) -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Clear
function Clear-anynodeNodeTransportConnections {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $RemoveRoutes,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)
	
	$Parameters=@{node=$Node}
	if ($RemoveRoutes -ne $null -AND $RemoveRoutes -ne '') {
		$Parameters.Add('removeRoutes',$RemoveRoutes)
	}

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /nodes/transportConnections/clear -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Transport Connections
#	Get
function Get-anynodeTransportConnection {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Id,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /transportConnections/get -Parameters @{id=$Id} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get Authentication
function Get-anynodeTransportConnectionAuthentication {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $TransportConnection,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /transportConnections/authentication/get -Parameters @{transportConnection=$TransportConnection} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set Authentication
function Set-anynodeTransportConnectionAuthentication {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $TransportConnection,
		[Parameter()]
		[string] $TransportConnectionUsername,
		[Parameter()]
		[string] $TransportConnectionPassword,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}
	if (!($TransportConnectionUsername -eq '')) {
		$Body.Add('username',$TransportConnectionUsername)
	}

	if (!($TransportConnectionPassword -eq '')) {
		$Body.Add('password',$TransportConnectionPassword)
	}

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /transportConnections/authentication/set -Body ($Body | ConvertTo-Json -Depth 99) -Parameters @{transportConnection=$TransportConnection} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get Asserted URI
function Get-anynodeTransportConnectionAssertedURI {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $TransportConnection,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /transportConnections/assertedUri/get -Parameters @{transportConnection=$TransportConnection} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set Asserted URI
function Set-anynodeTransportConnectionAssertedURI {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $TransportConnection,
		[Parameter()]
		[string] $URI,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}
	if (!($URI -eq '')) {
		$Body.Add('iri',$URI)
	}

	if (!($DisplayName -eq '')) {
		$Body.Add('displayName',$DisplayName)
	}

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /transportConnections/assertedUri/set -Body ($Body | ConvertTo-Json -Depth 99) -Parameters @{transportConnection=$TransportConnection} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get Address of Record
function Get-anynodeTransportConnectionAddressOfRecord {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $TransportConnection,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /transportConnections/addressOfRecord/get -Parameters @{transportConnection=$TransportConnection} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set Address of Record
function Set-anynodeTransportConnectionAddressOfRecord {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $TransportConnection,
		[Parameter()]
		[string] $URI,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}
	if (!($URI -eq '')) {
		$Body.Add('iri',$URI)
	}

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /transportConnections/addressOfRecord/set -Body ($Body | ConvertTo-Json -Depth 99) -Parameters @{transportConnection=$TransportConnection} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add
#		Standard
function Add-anynodeStandardTransportConnection {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $DisplayName,
		[Parameter()]
		[string] $SipNodeId,
		[Parameter()]
		[ValidateSet($true, $false)]
		[psobject] $UseAProxyServer,
		[Parameter()]
		[psobject] $Proxy,
		[Parameter()]
		[ValidateSet($true, $false)]
		[psobject] $RegistrationEnabled,
		[Parameter()]
		[string] $SipDomain,
		[Parameter()]
		[string] $AddressOfRecord,
		[Parameter()]
		[ValidateSet($true, $false)]
		[psobject] $AuthenticationEnabled,
		[Parameter()]
		[psobject] $AuthenticationPolicy,
		[Parameter()]
		[string] $AssertedURI,
		[Parameter()]
		[string] $AssertedDisplayName,
		[Parameter()]
		[string] $ConditionId,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}
	if (!($SipNodeId -eq '')) {
		$Body.Add('sipNodeId',$SipNodeId)
	}

	if (!($Proxy -eq $null)) {
		$Body.Add('proxy',$Proxy)
	}

	if ($RegistrationEnabled) {
		$Registration=@{}
		if (!($SipDomain -eq '')) {
			$Registration.Add('sipDomain',$SipDomain)
		}

		if (!($AddressOfRecord -eq '')) {
			$Registration.Add('addressOfRecord',$AddressOfRecord)
		}

		$Body.Add('registration',$Registration)
	}

	if (!($AuthenticationPolicy -eq '')) {
		$Body.Add('credentials',$AuthenticationPolicy)
	}

	if (!($AssertedURI -eq '')) {
		$AssertedAddress=@{iri=$AssertedURI}

		if (!($AssertedDisplayName -eq '')) {
			$AssertedAddress.Add('displayName',$AssertedDisplayName)
		}

		$Body.Add('assertedAddress',$AssertedAddress)
	}

	if (!($ConditionId -eq '')) {
		$Body.Add('conditionId',$ConditionId)
	}

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /transportConnections/addStandard -Body ($Body | ConvertTo-Json -Depth 99) -Parameters @{displayName=$DisplayName} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Object
#	Get
#		Single
function Get-anynodeObject {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Id,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{id=$Id}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /objects/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#		Multiple
function Get-anynodeObjects {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $Sort,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if ($Sort) {
		$Parameters.Add("sort",$Sort)
	}

	if ($DisplayName -ne '') {
		$Parameters.Add("displayName",$DisplayName)
	}

	if ($Parameters.Sort -eq $null -AND $Parameters.DisplayName -eq $null) {
        do {
		    $p1=Read-Host 'Choose a filter: sort (s)/display name (d)/both (b)'
        } while ($p1 -eq $null)

        switch ($p1) {
            's'{
                $Parameters.Add('sort',(Read-Host 'Sort'))
            }
            'd' {
                $Parameters.Add('displayName',(Read-Host 'Display Name'))
            }
            'b' {
                $Parameters.Add('sort',(Read-Host 'Sort'))
                $Parameters.Add('displayName',(Read-Host 'Display Name'))
            }
            default {
                echo 'Invalid input. Needs to be "s" or "d" or "b".'
                return
            }
        }
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /objects/sorts -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set
function Set-anynodeObject {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Id,
		[Parameter(mandatory=$true)]
		[psobject] $Object,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=($Object | ConvertTo-Json -Depth 99)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /objects/set -Body $Body -Parameters @{id=$Id} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Remove
function Remove-anynodeObject {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Id,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{id=$Id}

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /objects/remove -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Import
function Import-anynodeObject {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Filepath,
		[Parameter()]
		[ValidateSet("SKIP","OVERRIDE","CLONE","CANCEL")]
		[psobject] $DuplicateOption,
		[Parameter()]
		[string] $CloneComment,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $ReassignPortOnCollision,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if (!($DuplicateOption -eq $null)) {
		$Parameters.Add("action",$DuplicateOption)
	}

	if (!($CloneComment -eq '')) {
		$Parameters.Add("cloneComment",$CloneComment)
	}

	if ($ReassignPortOnCollision -ne $null) {
		$Parameters.Add("reassignPortOnCollision",$ReassignPortOnCollision)
	}

	Invoke-anynodeUpload -Address $Address -Port $Port -URITail /objects/import -Filenames @($Filepath) -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Export
function Export-anynodeObject {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Id,
		[Parameter()]
		[string] $Filename,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludePrivateData,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{id=$Id}
	if (!($IncludePrivateData -eq $null)) {
		$Parameters.Add("includePrivateData",$IncludePrivateData)
	}

	Invoke-anynodeDownload -Address $Address -Port $Port -URITail /objects/export -Filename $Filename -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Resources
#	Get
function Get-anynodeResources {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[ValidateSet('disk','cpu','ram')]
		[string] $Value,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if ($Value -ne $null -AND $Value -ne '') {
		$Parameters.Add("value",$Value)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /resources/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}
# Routing Domain
#	Get
function Get-anynodeRoutingDomains {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /routingDomains/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Proceeding Signal
#		Get
function Get-anynodeRoutingDomainProceedingSignal {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter(mandatory=$true)]
		[string] $RoutingDomain,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}

	if ($RoutingDomain -ne $null) {
		$Parameters.Add("routingDomain",$RoutingDomain)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /routingDomains/proceedingSignal/get -Credential $Credential -Parameters $Parameters -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#		Enable
function Enable-anynodeRoutingDomainProceedingSignal {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter(mandatory=$true)]
		[string] $RoutingDomain,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}

	if ($RoutingDomain -ne $null) {
		$Parameters.Add("routingDomain",$RoutingDomain)
	}

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /routingDomains/proceedingSignal/enable -Credential $Credential -Parameters $Parameters -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#		Disable
function Disable-anynodeRoutingDomainProceedingSignal {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter(mandatory=$true)]
		[string] $RoutingDomain,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}

	if ($RoutingDomain -ne $null) {
		$Parameters.Add("routingDomain",$RoutingDomain)
	}

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /routingDomains/proceedingSignal/disable -Credential $Credential -Parameters $Parameters -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Route in Maintenance
#		Get
function Get-anynodeRoutingDomainRouteInMaintenance {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter(mandatory=$true)]
		[string] $RoutingDomain,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}

	if ($RoutingDomain -ne $null) {
		$Parameters.Add("routingDomain",$RoutingDomain)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /routingDomains/routeInMaintenance/get -Credential $Credential -Parameters $Parameters -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#		Enable
function Enable-anynodeRoutingDomainRouteInMaintenance {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter(mandatory=$true)]
		[string] $RoutingDomain,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}

	if ($RoutingDomain -ne $null) {
		$Parameters.Add("routingDomain",$RoutingDomain)
	}

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /routingDomains/routeInMaintenance/enable -Credential $Credential -Parameters $Parameters -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#		Disable
function Disable-anynodeRoutingDomainRouteInMaintenance {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter(mandatory=$true)]
		[string] $RoutingDomain,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}

	if ($RoutingDomain -ne $null) {
		$Parameters.Add("routingDomain",$RoutingDomain)
	}

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /routingDomains/routeInMaintenance/disable -Credential $Credential -Parameters $Parameters -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add
function Add-anynodeSimpleRoutingDomain {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $DisplayName,
		[Parameter(mandatory=$true)]
		[string] $Node1,
		[Parameter(mandatory=$true)]
		[string] $Node2,
		[Parameter()]
		[string] $PrefixTo1,
		[Parameter()]
		[string] $PrefixTo2,
		[Parameter()]
		[string] $RoutingForwardProfile,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion=0,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{"node1"=$Node1}
	$Body.Add("node2",$Node2)

	if (!($PrefixTo1 -eq '')) {
		$Body.Add("prefixTo1",$PrefixTo1)
	}

	if (!($PrefixTo2 -eq '')) {
		$Body.Add("prefixTo2",$PrefixTo2)
	}

	if (!($RoutingForwardProfile -eq '')) {
		$Body.Add("routingForwardProfile",$RoutingForwardProfile)
	}

	if (!($DisplayName -eq '')) {
		$Body.Add("displayName",$DisplayName)
	}

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /routingDomains/add -Body ($Body | ConvertTo-Json) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

function Add-anynodeRoutingDomain {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[Object[]] $SourceNodes,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion=1,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}
	$Body.Add("displayName",$DisplayName)
    $Body.Add("sourceNodes",$SourceNodes)

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /routingDomains/add -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

function Set-anynodeRoutingDomain {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[Object[]] $SourceNodes,
		[Parameter(mandatory=$true)]
		[string] $RoutingDomain,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion=0,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}
	$Body.Add("displayName",$DisplayName)
    $Body.Add("sourceNodes",$SourceNodes)

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /routingDomains/set -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#   Remove
function Remove-anynodeRoutingDomain {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $RoutingDomain,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /routingDomains/remove -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password -Parameters @{routingDomain=$RoutingDomain}
}

# Source Nodes
#	Get
function Get-anynodeRoutingDomainSourceNodes {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $RoutingDomain,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /routingDomains/sourceNodes/get -Parameters @{routingDomain=$RoutingDomain} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add
function Add-anynodeRoutingDomainSourceNodes {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $RoutingDomain,
		[Parameter()]
		[Object[]] $SourceNodeIds,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=($SourceNodeIds | ConvertTo-Json)

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /routingDomains/sourceNodes/add -Parameters @{routingDomain=$RoutingDomain} -Body $Body -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set
function Set-anynodeRoutingDomainSourceNodes {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $RoutingDomain,
		[Parameter()]
		[Object[]] $SourceNodeIds,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=($SourceNodeIds | ConvertTo-Json)

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /routingDomains/sourceNodes/set -Parameters @{routingDomain=$RoutingDomain} -Body $Body -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Remove
function Remove-anynodeRoutingDomainSourceNodes {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $RoutingDomain,
		[Parameter()]
		[Object[]] $SourceNodeIds,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion=1,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	$Parameters.Add('routingDomain',$RoutingDomain)
	$Parameters.Add('nodeId',$SourceNodeIds)

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /routingDomains/sourceNodes/remove -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Routes
#	Get
function Get-anynodeRoutes {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
        [Parameter()]
		[ValidateSet('establishment','pathReplacement')]
		[string] $Mode,
		[Parameter()]
		[PSCredential] $Credential,
        [Parameter(mandatory=$true)]
		[string] $RoutingDomain,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

    $Parameters=@{routingDomain=$RoutingDomain}
    if ($Mode -ne $null -AND $Mode -ne '') {
		$Parameters.Add("mode",$Mode)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /routingDomains/routes/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Copy (Clone)
function Copy-anynodeRoute {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
        [Parameter()]
		[ValidateSet('establishment','pathReplacement')]
		[string] $Mode,
		[Parameter()]
		[PSCredential] $Credential,
        [Parameter(mandatory=$true)]
		[string] $RoutingDomain,
        [Parameter()]
		[string] $Index,
        [Parameter()]
		[string] $DisplayName,
        [Parameter()]
		[string] $Id,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

    $Parameters=@{routingDomain=$RoutingDomain}
    if ($Mode -ne $null -AND $Mode -ne '') {
		$Parameters.Add("mode",$Mode)
	}

	if ($Index -ne $null -AND $Index -ne '') {
		$Parameters.Add("index",$Index)
	}

	if ($DisplayName -ne $null -AND $DisplayName -ne '') {
		$Parameters.Add("displayName",$DisplayName)
	}

	if ($Id -ne $null -AND $Id -ne '') {
		$Parameters.Add("id",$Id)
	}

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /routingDomains/routes/clone -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add
function Add-anynodeEstablishmentRoute {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $RoutingDomain,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $UseUnconditionalRouting,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $UseRouteSupervision,
		[Parameter()]
		[string] $SupervisionId,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[Object[]] $Filters,
		[Parameter()]
		[ValidateSet('ROUTE','IGNORE','REJECT','REDIRECT','ESTABLISH_PARALLEL')]
		[string] $Type,
		[Parameter()]
		[string] $DestinationNodeId,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $SkipIfNodeNotOperational,
		[Parameter()]
		[ValidateSet('CREATE_NEW','USE_REFERENCE','NONE')]
		[string] $RoutingForwardProfileSelection,
		[Parameter()]
		[string] $RoutingForwardProfileDisplayName,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $AdjustRoutingForwardProfileSettings,
		[Parameter()]
		[ValidateSet('PASSTHROUGH','TRANSCODING','BYPASS','PASSTHROUGH_WITH_FALLBACK_TO_BYPASS')]
		[string] $RoutingForwardProfileMode,
		[Parameter()]
        [string] $RoutingForwardProfileId,
        [Parameter()]
        [ValidateSet('SUCCESS','DIAL_STRING','NETWORK_PERMISSION','NETWORK_CONGESTION','NETWORK_EQUIPMENT','USER_BUSY','USER_REDIRECTED','USER_NOT_RESPONDING','USER_NOT_SELECTED','USER_REJECTED','USER_TERMINATED','MEDIA_NEGOTIATION','ERROR','DOMAIN_SPECIFIC_0','DOMAIN_SPECIFIC_1','DOMAIN_SPECIFIC_2','DOMAIN_SPECIFIC_3','DOMAIN_SPECIFIC_4','DOMAIN_SPECIFIC_5','DOMAIN_SPECIFIC_6','DOMAIN_SPECIFIC_7','DOMAIN_SPECIFIC_8','DOMAIN_SPECIFIC_9','DOMAIN_SPECIFIC_10','DOMAIN_SPECIFIC_11','DOMAIN_SPECIFIC_12','DOMAIN_SPECIFIC_13','DOMAIN_SPECIFIC_14','DOMAIN_SPECIFIC_15','DOMAIN_SPECIFIC_16','DOMAIN_SPECIFIC_17','DOMAIN_SPECIFIC_18','DOMAIN_SPECIFIC_19')]
        [string] $RejectStatus,
        [Parameter()]
        [string] $ParallelCallDialStringDirectoryId,
        [Parameter()]
        [PSCustomObject] $SourceDialStringRewriting,
        [Parameter()]
        [PSCustomObject] $DestinationDialStringRewriting,
        [Parameter()]
        [Object[]] $ParallelCalls,
        [Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

    $Parameters=@{}
    $Parameters.Add("routingDomain",$RoutingDomain)
    $Parameters.Add("mode","establishment")

	$Body=@{}

    if (!($UseUnconditionalRouting -eq $null)) {
        $Body.Add("useUnconditionalRouting",$UseUnconditionalRouting)
    }

    if (!($UseRouteSupervision -eq $null)) {
        $Body.Add("useRouteSupervision",$UseRouteSupervision)
    }

    if (!($SupervisionId -eq $null)) {
        $Body.Add("supervisionId",$SupervisionId)
    }
    
    if (!($DisplayName -eq $null)) {
        $Body.Add("displayName",$DisplayName)
    }

    if (!($Filters -eq $null)) {
        $Body.Add("filters",$Filters)
    }
    
    $Establishment=@{}

    if (!($Type -eq $null)) {
        $Establishment.Add("type",$Type)
    }

    if (!($DestinationNodeId -eq $null)) {
        $Establishment.Add("destinationNodeId",$DestinationNodeId)
    }

    if (!($SkipIfNodeNotOperational -eq $null)) {
        $Establishment.Add("skipIfNodeNotOperational",$SkipIfNodeNotOperational)
    }

    if (!($RoutingForwardProfileSelection -eq $null)) {
        $Establishment.Add("routingForwardProfileSelection",$RoutingForwardProfileSelection)
    }

    if (!($RoutingForwardProfileDisplayName -eq $null)) {
        $Establishment.Add("routingForwardProfileDisplayName",$RoutingForwardProfileDisplayName)
    }

    if (!($AdjustRoutingForwardProfileSettings -eq $null)) {
        $Establishment.Add("adjustRoutingForwardProfileSettings",$AdjustRoutingForwardProfileSettings)
    }

    if (!($RoutingForwardProfileMode -eq $null)) {
        $Establishment.Add("routingForwardProfileMode",$RoutingForwardProfileMode)
    }

    if (!($RoutingForwardProfileId -eq $null)) {
        $Establishment.Add("routingForwardProfileId",$RoutingForwardProfileId)
    }

    if (!($RejectStatus -eq $null)) {
        $Establishment.Add("rejectStatus",$RejectStatus)
    }

    if (!($ParallelCallDialStringDirectoryId -eq $null)) {
        $Establishment.Add("parallelCallDialStringDirectoryId",$ParallelCallDialStringDirectoryId)
    }

    if (!($SourceDialStringRewriting -eq $null)) {
        $Establishment.Add("sourceDialStringRewriting",$SourceDialStringRewriting)
    }

    if (!($DestinationDialStringRewriting -eq $null)) {
        $Establishment.Add("destinationDialStringRewriting",$DestinationDialStringRewriting)
    }

    if (!($ParallelCalls -eq $null)) {
        $Establishment.Add("parallelCalls",$ParallelCalls)
    }

    $Body.Add("establishment",$Establishment)

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /routingDomains/routes/add -Parameters $Parameters -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

function Add-anynodePathReplacementRoute {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $RoutingDomain,
		[Parameter()]
		[ValidateSet($true,$false)]
        [pscustomobject] $UseUnconditionalRouting,
		[Parameter()]
		[ValidateSet($true,$false)]
        [pscustomobject] $UseRouteSupervision,
        [Parameter()]
        [string] $SupervisionId,
        [Parameter()]
        [string] $DisplayName,
        [Parameter()]
        [Object[]] $Filters,
        [Parameter()]
        [ValidateSet('REPLACE','REJECT')]
        [string] $Type,
        [Parameter()]
        [string] $DestinationNodeId,
        [Parameter()]
        [ValidateSet($true,$false)]
        [pscustomobject] $SkipIfNodeNotOperational,
        [Parameter()]
        [ValidateSet('CREATE_NEW','USE_REFERENCE','NONE')]
        [string] $RoutingForwardProfileSelection,
        [Parameter()]
        [string] $RoutingForwardProfileDisplayName,
        [Parameter()]
        [ValidateSet($true,$false)]
        [pscustomobject] $AdjustRoutingForwardProfileSettings,
        [Parameter()]
        [ValidateSet('PASSTHROUGH','TRANSCODING','BYPASS','PASSTHROUGH_WITH_FALLBACK_TO_BYPASS')]
        [string] $RoutingForwardProfileMode,
        [Parameter()]
        [string] $RoutingForwardProfileId,
        [Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

    $Parameters=@{}
    $Parameters.Add("routingDomain",$RoutingDomain)
    $Parameters.Add("mode","pathReplacement")

	$Body=@{}

    if (!($UseUnconditionalRouting -eq $null)) {
        $Body.Add("useUnconditionalRouting",$UseUnconditionalRouting)
    }

    if (!($UseRouteSupervision -eq $null)) {
        $Body.Add("useRouteSupervision",$UseRouteSupervision)
    }

    if (!($SupervisionId -eq $null)) {
        $Body.Add("supervisionId",$SupervisionId)
    }
    
    if (!($DisplayName -eq $null)) {
        $Body.Add("displayName",$DisplayName)
    }

    if (!($Filters -eq $null)) {
        $Body.Add("filters",$Filters)
    }

    $PathReplacement=@{}

    if (!($Type -eq $null)) {
        $PathReplacement.Add("type",$Type)
    }

    if (!($RoutingForwardProfileSelection -eq $null)) {
        $PathReplacement.Add("routingForwardProfileSelection",$RoutingForwardProfileSelection)
    }

    if (!($RoutingForwardProfileDisplayName -eq $null)) {
        $PathReplacement.Add("routingForwardProfileDisplayName",$RoutingForwardProfileDisplayName)
    }

    if (!($AdjustRoutingForwardProfileSettings -eq $null)) {
        $PathReplacement.Add("adjustRoutingForwardProfileSettings",$AdjustRoutingForwardProfileSettings)
    }

    if (!($RoutingForwardProfileMode -eq $null)) {
        $PathReplacement.Add("routingForwardProfileMode",$RoutingForwardProfileMode)
    }

    if (!($RoutingForwardProfileId -eq $null)) {
        $PathReplacement.Add("routingForwardProfileId",$RoutingForwardProfileId)
    }

    $Body.Add("pathReplacement",$PathReplacement)

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /routingDomains/routes/add -Parameters $Parameters -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set
function Set-anynodeEstablishmentRoute {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $RoutingDomain,
        [Parameter()]
        [string] $Index,
        [Parameter()]
        [string] $RouteName,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $UseUnconditionalRouting,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $UseRouteSupervision,
		[Parameter()]
		[string] $SupervisionId,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[Object[]] $Filters,
		[Parameter()]
		[ValidateSet('ROUTE','IGNORE','REJECT','REDIRECT','ESTABLISH_PARALLEL')]
		[string] $Type,
		[Parameter()]
		[string] $DestinationNodeId,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $SkipIfNodeNotOperational,
		[Parameter()]
		[ValidateSet('CREATE_NEW','USE_REFERENCE','NONE')]
		[string] $RoutingForwardProfileSelection,
		[Parameter()]
		[string] $RoutingForwardProfileDisplayName,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $AdjustRoutingForwardProfileSettings,
		[Parameter()]
		[ValidateSet('PASSTHROUGH','TRANSCODING','BYPASS','PASSTHROUGH_WITH_FALLBACK_TO_BYPASS')]
		[string] $RoutingForwardProfileMode,
		[Parameter()]
        [string] $RoutingForwardProfileId,
        [Parameter()]
        [ValidateSet('SUCCESS','DIAL_STRING','NETWORK_PERMISSION','NETWORK_CONGESTION','NETWORK_EQUIPMENT','USER_BUSY','USER_REDIRECTED','USER_NOT_RESPONDING','USER_NOT_SELECTED','USER_REJECTED','USER_TERMINATED','MEDIA_NEGOTIATION','ERROR','DOMAIN_SPECIFIC_0','DOMAIN_SPECIFIC_1','DOMAIN_SPECIFIC_2','DOMAIN_SPECIFIC_3','DOMAIN_SPECIFIC_4','DOMAIN_SPECIFIC_5','DOMAIN_SPECIFIC_6','DOMAIN_SPECIFIC_7','DOMAIN_SPECIFIC_8','DOMAIN_SPECIFIC_9','DOMAIN_SPECIFIC_10','DOMAIN_SPECIFIC_11','DOMAIN_SPECIFIC_12','DOMAIN_SPECIFIC_13','DOMAIN_SPECIFIC_14','DOMAIN_SPECIFIC_15','DOMAIN_SPECIFIC_16','DOMAIN_SPECIFIC_17','DOMAIN_SPECIFIC_18','DOMAIN_SPECIFIC_19')]
        [string] $RejectStatus,
        [Parameter()]
        [string] $ParallelCallDialStringDirectoryId,
        [Parameter()]
        [PSCustomObject] $SourceDialStringRewriting,
        [Parameter()]
        [PSCustomObject] $DestinationDialStringRewriting,
        [Parameter()]
        [Object[]] $ParallelCalls,
        [Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

    $Parameters=@{}
    $Parameters.Add("routingDomain",$RoutingDomain)
    $Parameters.Add("mode","establishment")

       if (!($Index -eq $null)) {
        $Parameters.Add("index",$Index)
    }

    if (!($RouteName -eq $null)) {
        $Parameters.Add("routeName",$RouteName)
    }

	$Body=@{}

    if (!($UseUnconditionalRouting -eq $null)) {
        $Body.Add("useUnconditionalRouting",$UseUnconditionalRouting)
    }

    if (!($UseRouteSupervision -eq $null)) {
        $Body.Add("useRouteSupervision",$UseRouteSupervision)
    }

    if (!($SupervisionId -eq $null)) {
        $Body.Add("supervisionId",$SupervisionId)
    }
    
    if (!($DisplayName -eq $null)) {
        $Body.Add("displayName",$DisplayName)
    }

    if (!($Filters -eq $null)) {
        $Body.Add("filters",$Filters)
    }
    
    $Establishment=@{}

    if (!($Type -eq $null)) {
        $Establishment.Add("type",$Type)
    }

    if (!($DestinationNodeId -eq $null)) {
        $Establishment.Add("destinationNodeId",$DestinationNodeId)
    }

    if (!($SkipIfNodeNotOperational -eq $null)) {
        $Establishment.Add("skipIfNodeNotOperational",$SkipIfNodeNotOperational)
    }

    if (!($RoutingForwardProfileSelection -eq $null)) {
        $Establishment.Add("routingForwardProfileSelection",$RoutingForwardProfileSelection)
    }

    if (!($RoutingForwardProfileDisplayName -eq $null)) {
        $Establishment.Add("routingForwardProfileDisplayName",$RoutingForwardProfileDisplayName)
    }

    if (!($AdjustRoutingForwardProfileSettings -eq $null)) {
        $Establishment.Add("adjustRoutingForwardProfileSettings",$AdjustRoutingForwardProfileSettings)
    }

    if (!($RoutingForwardProfileMode -eq $null)) {
        $Establishment.Add("routingForwardProfileMode",$RoutingForwardProfileMode)
    }

    if (!($RoutingForwardProfileId -eq $null)) {
        $Establishment.Add("routingForwardProfileId",$RoutingForwardProfileId)
    }

    if (!($RejectStatus -eq $null)) {
        $Establishment.Add("rejectStatus",$RejectStatus)
    }

    if (!($ParallelCallDialStringDirectoryId -eq $null)) {
        $Establishment.Add("parallelCallDialStringDirectoryId",$ParallelCallDialStringDirectoryId)
    }

    if (!($SourceDialStringRewriting -eq $null)) {
        $Establishment.Add("sourceDialStringRewriting",$SourceDialStringRewriting)
    }

    if (!($DestinationDialStringRewriting -eq $null)) {
        $Establishment.Add("destinationDialStringRewriting",$DestinationDialStringRewriting)
    }

    if (!($ParallelCalls -eq $null)) {
        $Establishment.Add("parallelCalls",$ParallelCalls)
    }

    $Body.Add("establishment",$Establishment)

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /routingDomains/routes/set -Parameters $Parameters -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

function Set-anynodePathReplacementRoute {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $RoutingDomain,
        [Parameter()]
        [string] $Index,
        [Parameter()]
        [string] $RouteName,
		[Parameter()]
		[ValidateSet($true,$false)]
        [pscustomobject] $UseUnconditionalRouting,
		[Parameter()]
		[ValidateSet($true,$false)]
        [pscustomobject] $UseRouteSupervision,
        [Parameter()]
        [string] $SupervisionId,
        [Parameter()]
        [string] $DisplayName,
        [Parameter()]
        [Object[]] $Filters,
        [Parameter()]
        [ValidateSet('ACCEPT','REJECT')]
        [string] $Type,
        [Parameter()]
        [string] $DestinationNodeId,
        [Parameter()]
        [ValidateSet($true,$false)]
        [pscustomobject] $SkipIfNodeNotOperational,
        [Parameter()]
        [ValidateSet('CREATE_NEW','USE_REFERENCE','NONE')]
        [string] $RoutingForwardProfileSelection,
        [Parameter()]
        [string] $RoutingForwardProfileDisplayName,
        [Parameter()]
        [ValidateSet($true,$false)]
        [pscustomobject] $AdjustRoutingForwardProfileSettings,
        [Parameter()]
        [ValidateSet('PASSTHROUGH','TRANSCODING','BYPASS','PASSTHROUGH_WITH_FALLBACK_TO_BYPASS')]
        [string] $RoutingForwardProfileMode,
        [Parameter()]
        [string] $RoutingForwardProfileId,
        [Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

    $Parameters=@{}
    $Parameters.Add("routingDomain",$RoutingDomain)
    $Parameters.Add("mode","pathReplacement")

      if (!($Index -eq $null)) {
        $Parameters.Add("index",$Index)
    }

    if (!($RouteName -eq $null)) {
        $Parameters.Add("routeName",$RouteName)
    }


	$Body=@{}

    if (!($UseUnconditionalRouting -eq $null)) {
        $Body.Add("useUnconditionalRouting",$UseUnconditionalRouting)
    }

    if (!($UseRouteSupervision -eq $null)) {
        $Body.Add("useRouteSupervision",$UseRouteSupervision)
    }

    if (!($SupervisionId -eq $null)) {
        $Body.Add("supervisionId",$SupervisionId)
    }
    
    if (!($DisplayName -eq $null)) {
        $Body.Add("displayName",$DisplayName)
    }

    if (!($Filters -eq $null)) {
        $Body.Add("filters",$Filters)
    }

    $PathReplacement=@{}

    if (!($Type -eq $null)) {
        $PathReplacement.Add("type",$Type)
    }

    if (!($RoutingForwardProfileSelection -eq $null)) {
        $PathReplacement.Add("routingForwardProfileSelection",$RoutingForwardProfileSelection)
    }

    if (!($RoutingForwardProfileDisplayName -eq $null)) {
        $PathReplacement.Add("routingForwardProfileDisplayName",$RoutingForwardProfileDisplayName)
    }

    if (!($AdjustRoutingForwardProfileSettings -eq $null)) {
        $PathReplacement.Add("adjustRoutingForwardProfileSettings",$AdjustRoutingForwardProfileSettings)
    }

    if (!($RoutingForwardProfileMode -eq $null)) {
        $PathReplacement.Add("routingForwardProfileMode",$RoutingForwardProfileMode)
    }

    if (!($RoutingForwardProfileId -eq $null)) {
        $PathReplacement.Add("routingForwardProfileId",$RoutingForwardProfileId)
    }

    $Body.Add("pathReplacement",$PathReplacement)

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /routingDomains/routes/set -Parameters $Parameters -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Remove
function Remove-anynodeEstablishmentRoute {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $RoutingDomain,
		[Parameter()]
		[int] $Index=-1,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{routingDomain=$RoutingDomain}
	$Parameters.Add("mode","establishment")

	if ($Index -gt -1) {
		$Parameters.Add("index",$Index)
	} elseif ($DisplayName -ne $null -AND $DisplayName -ne '') {
		$Parameters.Add("displayName",$DisplayName)
	}

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /routingDomains/routes/remove -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

function Remove-anynodePathReplacementRoute {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $RoutingDomain,
		[Parameter()]
		[int] $Index=-1,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{routingDomain=$RoutingDomain}
	$Parameters.Add("mode","pathReplacement")

	if ($Index -gt -1) {
		$Parameters.Add("index",$Index)
	} elseif ($DisplayName -ne $null -AND $DisplayName -ne '') {
		$Parameters.Add("displayName",$DisplayName)
	}

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /routingDomains/routes/remove -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Routing Forward Profile
#	Get
function Get-anynodeRoutingForwardProfiles {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[switch] $CompleteData,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /routingForwardProfiles/get -Parameters @{completeData=$CompleteData} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add
function Add-anynodeRoutingForwardProfile {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $DisplayName,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $RemoveAnynodeFromSignalingPath,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $EarlyMediaPassThrough,
		[Parameter()]
		[pscustomobject] $TelephonyForwardingStack,
		[Parameter()]
		[string] $TelephonyForwardingStackId,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $RecordIncoming,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $RecordOutgoing,
		[Parameter()]
		[string] $MediaRecorderId,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $AnnouncementIncoming,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $AnnouncementOutgoing,
		[Parameter()]
		[string] $AnnouncementMediaSourceId,
		[Parameter()]
		[string] $MediaNegotiationForwarderId,
		[Parameter()]
		[PSCustomObject] $SignalingForwardingFromCallingToCalled,
		[Parameter()]
		[PSCustomObject] $SignalingForwardingFromCalledToCalling,
		[Parameter()]
		[string] $MediaTranscodingOptionsId,
		[Parameter()]
		[ValidateSet('TRANSCODING','BYPASS','PASSTHROUGH','PASSTHROUGH_WITH_FALLBACK_TO_BYPASS')]
		[string] $MediaMode,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}

	$Body.Add('displayName',$DisplayName)

	if (!($RemoveAnynodeFromSignalingPath -eq $null)) {
		$Body.Add('removeAnynodeFromSignalingPath',$RemoveAnynodeFromSignalingPath)
	}

	if (!($EarlyMediaPassThrough -eq $null)) {
		$Body.Add('earlyMediaPassThrough',$EarlyMediaPassThrough)
	}

	if (!($TelephonyForwardingStack -eq $null)) {
		$Body.Add('telephonyForwardingStack',$TelephonyForwardingStack)
	}

	if (!($TelephonyForwardingStackId -eq '')) {
		$Body.Add('telephonyForwardingStackId',$TelephonyForwardingStackId)
	}

	if (!($RecordIncoming -eq $null)) {
		$Body.Add('recordIncoming',$RecordIncoming)
	}

	if (!($RecordOutgoing -eq $null)) {
		$Body.Add('recordOutgoing',$RecordOutgoing)
	}

	if (!($MediaRecorderId -eq '')) {
		$Body.Add('mediaRecorderId',$MediaRecorderId)
	}

	if (!($AnnouncementIncoming -eq $null)) {
		$Body.Add('announcementIncoming',$AnnouncementIncoming)
	}

	if (!($AnnouncementOutgoing -eq $null)) {
		$Body.Add('announcementOutgoing',$AnnouncementOutgoing)
	}

	if (!($AnnouncementMediaSourceId -eq '')) {
		$Body.Add('announcementMediaSourceId',$AnnouncementMediaSourceId)
	}

	if (!($MediaNegotiationForwarderId -eq '')) {
		$Body.Add('mediaNegotiationForwarderId',$MediaNegotiationForwarderId)
	}

	if (!($SignalingForwardingFromCallingToCalled -eq $null)) {
		$Body.Add('signalingForwardingFromCallingToCalled',$SignalingForwardingFromCallingToCalled)
	}

	if (!($SignalingForwardingFromCalledToCalling -eq $null)) {
		$Body.Add('signalingForwardingFromCalledToCalling',$SignalingForwardingFromCalledToCalling)
	}

	if (!($MediaTranscodingOptionsId -eq '')) {
		$Body.Add('mediaTranscodingOptionsId',$MediaTranscodingOptionsId)
	}

	if (!($MediaMode -eq '')) {
		$Body.Add('mediaMode',$MediaMode)
	}

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /routingForwardProfiles/add -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Roles
#   Get
function Get-anynodeRoles {
    [CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

    Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /roles/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#   Add
function Add-anynodeRole {
    [CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
        [Parameter()]
        [string] $Id,
        [Parameter(Mandatory=$true)]
        [string] $DisplayName,
        [Parameter()]
        [ValidateSet('ACCESS','NO_ACCESS')]
        [string] $UserInterface,
        [Parameter()]
        [ValidateSet('ACCESS','NO_ACCESS')]
        [string] $API,
        [Parameter()]
        [ValidateSet('READ_ACCESS','READ_WRITE_ACCESS','NO_ACCESS')]
        [string] $UserManagement,
        [Parameter()]
        [ValidateSet('READ_ACCESS','READ_WRITE_ACCESS','NO_ACCESS')]
        [string] $WebServer,
        [Parameter()]
        [ValidateSet('READ_ACCESS','READ_WRITE_ACCESS','NO_ACCESS')]
        [string] $Backends,
        [Parameter()]
        [ValidateSet('ACCESS','NO_ACCESS')]
        [string] $Administration,
        [Parameter()]
        [ValidateSet('READ_ACCESS','READ_WRITE_ACCESS','NO_ACCESS')]
        [string] $MonitoringMode,
        [Parameter()]
        [ValidateSet('ACCESS','NO_ACCESS')]
        [string] $Tracing,
        [Parameter()]
        [ValidateSet('READ_ACCESS','READ_WRITE_ACCESS','NO_ACCESS')]
        [string] $ConfigurationMode,
        [Parameter()]
        [ValidateSet('ACCESS','NO_ACCESS')]
        [string] $MediaRecording,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

    $Body=@{}
	$Permissions=@{}

    if ($Id -eq $null -OR $Id -eq '') {
        $Id=New-Guid
    }

    $Body.Add('id',$Id)
    $Body.Add('displayName',$DisplayName)
	
    
    if ($UserInterface -ne $null) {
        $Permissions.Add('userInterface',$UserInterface)
    }

    if ($API -ne $null) {
        $Permissions.Add('api',$API)
    }

    if ($UserManagement -ne $null) {
        $Permissions.Add('userManagement',$UserManagement)
    }

    if ($WebServer -ne $null) {
        $Permissions.Add('webServer',$WebServer)
    }

    if ($Backends -ne $null) {
        $Permissions.Add('backends',$Backends)
    }

    if ($Administration -ne $null) {
        $Permissions.Add('administration',$Administration)
    }

    if ($MonitoringMode -ne $null) {
        $Permissions.Add('monitoringMode',$MonitoringMode)
    }

    if ($Tracing -ne $null) {
        $Permissions.Add('tracing',$Tracing)
    }

    if ($ConfigurationMode -ne $null) {
        $Permissions.Add('configurationMode',$ConfigurationMode)
    }

    if ($MediaRecording -ne $null) {
        $Permissions.Add('mediaRecording',$MediaRecording)
    }
	
	if ($Permissions -ne $null) {
        $Body.Add('permissions',$Permissions)
    }

    Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /roles/add -Body ($Body | ConvertTo-Json) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#   Remove
function Remove-anynodeRole {
    [CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
        [Parameter(Mandatory=$true)]
        [string] $Role,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

    $Parameters=@{role=$Role}

    Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /roles/remove -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Microsoft Teams SBA
#	Get
function Get-anynodeMsTeamsSba {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /sba/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set
function Set-anynodeMsTeamsSba {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $IpAddress,
		[Parameter()]
		[PSCustomObject] $Fingerprint,
		[Parameter()]
		[string] $Application,
		[Parameter()]
		[string] $Secret,
		[Parameter()]
		[string] $Identity,
		[Parameter()]
		[string] $TenantId,
		[Parameter()]
		[PSCustomObject] $Log,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}
	
	if ($IpAddress -ne $null) {
		$Body.Add("ipAddress", $IpAddress)
	}

	if ($Fingerprint -ne $null) {
		$Body.Add("fingerprint", $Fingerprint)
	}

	if ($Application -ne $null) {
		$Body.Add("application", $Application)
	}

	if ($Secret -ne $null) {
		$Body.Add("secret", $Secret)
	}

	if ($Identity -ne $null) {
		$Body.Add("identity", $Identity)
	}

	if ($TenantId -ne $null) {
		$Body.Add("tenantId", $TenantId)
	}

	if ($Log -ne $null) {
		$Body.Add("log", $Log)
	}

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /sba/set -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Enable
function Enable-anynodeMsTeamsSba {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /sba/enable -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Disable
function Disable-anynodeMsTeamsSba {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /sba/disable -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Users
#		Get
function Get-anynodeMsTeamsSbaUsers {
[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /sba/users/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#		Update
function Update-anynodeMsTeamsSbaUsers {
[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /sba/users/refresh -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Status
#		Get
function Get-anynodeMsTeamsSbaStatus {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /sba/status/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Private Key
#		Get
function Get-anynodeMsTeamsSbaPrivateKey {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $Filename,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeDownload -Address $Address -Port $Port -URITail /sba/privateKey/get -Filename $Filename -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#		Set
function Set-anynodeMsTeamsSbaPrivateKey {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Filepath,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeUpload -Address $Address -Port $Port -URITail /sba/privateKey/set -Filenames @($Filepath) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Certificate
#		Get
function Get-anynodeMsTeamsSbaCertificate {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $Filename,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeDownload -Address $Address -Port $Port -URITail /sba/certificate/get -Filename $Filename -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#		Set
function Set-anynodeMsTeamsSbaCertificate {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Filepath,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeUpload -Address $Address -Port $Port -URITail /sba/certificate/set -Filenames @($Filepath) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Fingerprint
#		Get
function Get-anynodeMsTeamsSbaFingerprint {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /sba/fingerprint/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#		Set
function Set-anynodeMsTeamsSbaFingerprint {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
        [ValidateSet('NONE','CUSTOM','MS_TEAMS_NODE')]
        [string] $Source,
		[Parameter()]
        [string] $Value,
		[Parameter()]
        [string] $MsTeamsNodeId,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}

	$Body.Add("source", $Source)

	if ($Value -ne $null) {
		$Body.Add("value", $Value)
	}

	if ($MsTeamsNodeId -ne $null) {
		$Body.Add("msTeamsNodeId", $MsTeamsNodeId)
	}

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /sba/fingerprint/set -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	IP Address
#		Get
function Get-anynodeMsTeamsSbaIpAddress {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /sba/ipAddress/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#		Set
function Set-anynodeMsTeamsSbaIpAddress {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $IpAddress,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /sba/ipAddress/set -Body $IpAddress -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Identity
#		Get
function Get-anynodeMsTeamsSbaIdentity {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /sba/identity/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#		Set
function Set-anynodeMsTeamsSbaIdentity {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Identity,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /sba/identity/set -Body $Identity -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Tenant ID
#		Get
function Get-anynodeMsTeamsSbaTenantId {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /sba/tenantId/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#		Set
function Set-anynodeMsTeamsSbaTenantId {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $TenantId,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /sba/tenantId/set -Body $TenantId -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Application
#		Get
function Get-anynodeMsTeamsSbaApplication {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /sba/application/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#		Set
function Set-anynodeMsTeamsSbaApplication {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Application,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /sba/application/set -Body $Application -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Secret
#		Set
function Set-anynodeMsTeamsSbaSecret {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[SecureString] $Secret,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /sba/secret/set -Body [System.Runtime.InteropServices.Marshal]::PtrToStringAuto([System.Runtime.InteropServices.Marshal]::SecureStringToBSTR($Secret)) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Log
#		Get
function Get-anynodeMsTeamsSbaLog {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /sba/log/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#		Set
function Set-anynodeMsTeamsSbaLog {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $Directory,
		[Parameter()]
        [ValidateSet('Trace','Debug','Info','Warn','Error','Fatal')]
		[string] $Level,
		[Parameter()]
		[int] $MaxArchiveFiles,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}

	if ($Directory -ne $null) {
		$Body.Add("directory", $Directory)
	}

	if ($Level -ne $null) {
		$Body.Add("level", $Level)
	}

	if ($MaxArchiveFiles -ne $null) {
		$Body.Add("maxArchiveFiles", $MaxArchiveFiles)
	}

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /sba/log/set -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Sorts
#	Get
function Get-anynodeSorts {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /objects/sorts -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# SystemIds
#	Get
function Get-anynodeSystemIds {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /licenses/systemIds/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Trace
#	Get
function Get-anynodeTrace {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $Filename,
		[Parameter()]
        [ValidateSet('anynode','anynodeFrontend','anynodeMonitor','anynodeAdministration')]
        [string] $Product='anynode',
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if (!($Product -eq '')) {
		$Parameters.Add("product",$Product)
	}

	Invoke-anynodeDownload -Address $Address -Port $Port -URITail /trace/get -Filename $Filename -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Tracing
#	Get
function Get-anynodeTracing {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[switch] $IncludeDefaults,
		[Parameter()]
        [ValidateSet('anynode','anynodeFrontend','anynodeMonitor','anynodeAdministration')]
        [string] $Product='anynode',
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if ($IncludeDefaults) {
		$Parameters.Add("includeDefaults",$true)
	}
	
	if (!($Product -eq '')) {
			$Parameters.Add("product",$Product)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /tracing/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set
function Set-anynodeTracing {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[ValidateSet('FILE','SPLIT')]
		[string] $Selection,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $TraceActive,
		[Parameter()]
		[pscustomobject] $Options,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludePrivateData,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludeDefaults,
		[Parameter()]
        [ValidateSet('anynode','anynodeFrontend','anynodeMonitor','anynodeAdministration')]
        [string] $Product='anynode',
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}

	if (!($Selection -eq '')) {
		$Body.Add("selection",$Selection)
	}

	if (!($TraceActive -eq $null)) {
		$Body.Add("traceActive",$TraceActive)
	}

	if (!($Options -eq $null)) {
		$Body.Add("options",($Options | ConvertTo-Json -Depth 99 | ConvertFrom-Json))
	}

	if (!($IncludePrivateData -eq $null)) {
		$Body.Add("includePrivateData",$IncludePrivateData)
	}

	if (!($IncludeDefaults -eq $null)) {
		$Body.Add("includeDefaults",$IncludeDefaults)
	}
	if (!($Product -eq '')) {
		$Body.Add("product",$Product)
	}

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /tracing/set -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Users
#   Get
function Get-anynodeUsers {
    [CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

    Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /users/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#   Add
function Add-anynodeUser {
    [CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
        [ValidateSet($true,$false)]
		[PSCustomObject] $LoginLocally=$true,
		[Parameter()]
        [ValidateSet($true,$false)]
		[PSCustomObject] $LdapLogin=$false,
        [Parameter(Mandatory=$true)]
        [string] $DisplayName,
        [Parameter()]
        [ValidateSet($true,$false)]
        [PSCustomObject] $Enabled=$true,
        [Parameter()]
        [string] $EmailAddress,
        [Parameter()]
        [string] $LocalUsername,
        [Parameter()]
        [string] $LocalPassword,
        [Parameter()]
        [PSCustomObject] $LdapAuthentications,
        [Parameter()]
        [ValidateSet('NOT_NECESSARY','REQUIRED','RECOMMENDED')]
        [string] $PasswordChange='REQUIRED',
        [Parameter()]
        [string[]] $RoleIds,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

    $Body=@{}

    $Body.Add('loginLocally',$LoginLocally)
    $Body.Add('ldapLogin',$LdapLogin)
    $Body.Add('displayName',$DisplayName)
    $Body.Add('enabled',$Enabled)

    if ($EmailAddress -ne $null -AND $EmailAddress -ne '') {
        $Body.Add('emailAddress',$EmailAddress)
    }

    if ($LocalUsername -ne $null -AND $LocalUsername -ne '') {
        $Body.Add('localUsername',$LocalUsername)
    }

    if ($LocalPassword -ne $null -AND $LocalPassword -ne '') {
        $Body.Add('localPassword',$LocalPassword)
    }

    if ($LdapAuthentications -ne $null) {
        $Body.Add('ldapAuthentications',$LdapAuthentications)
    }

    $Body.Add('passwordChange',$PasswordChange)

    if ($RoleIds -ne $null) {
        $Body.Add('roleIds',$RoleIds)
    }

    Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /users/add -Body ($Body | ConvertTo-Json) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#   Remove
function Remove-anynodeUser {
    [CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
        [Parameter()]
        [string] $User,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

    Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /users/remove -Parameters @{user=$User} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#    Change Password
function Submit-anynodePasswordChange {
    [CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
        [Parameter()]
        [string] $User,
        [Parameter(Mandatory=$true)]
        [string] $OldPassword,
        [Parameter(Mandatory=$true)]
        [string] $NewPassword,
        [Parameter()]
        [ValidateSet('NOT_NECESSARY','REQUIRED','RECOMMENDED')]
        [string] $PasswordChange,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

    $Parameters=@{}

    if ($User -ne $null -AND $User -ne '') {
        $Parameters.Add('user',$User)
    }

    $Parameters.Add('oldPassword',$OldPassword)
    $Parameters.Add('newPassword',$NewPassword)

    if ($PasswordChange -ne $null -AND $PasswordChange -ne 'NOT_NECESSARY') {
        $Parameters.Add('passwordChange',$PasswordChange)
    }

    Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /users/changePassword -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#    Get LDAP Authorities
function Get-anynodeLdapAuthorities {
    [CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

    Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /users/ldapAuthorities/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Version
#	Get
function Get-anynodeVersion {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[ValidateSet('anynode','java','ucma','administration','monitor','frontend')]
		[string] $Product,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if (!($Product -eq '')) {
		$Parameters.Add("product",$Product)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /version/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

function Create-anynodeConfigBackup {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[psobject] $IncludedData,
		[Parameter()]
		[ValidateSet($true,$false)]
		[bool] $EncryptData,
		[Parameter()]
		[string] $ZipPassword,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter(mandatory=$true)]
		[string] $Filename,
		[Parameter()]
		[int] $APIVersion=0,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}

	if ($IncludedData -ne $null) {
		$Body.add("includedData",$IncludedData)
	}

	if ($ZipPassword -ne $null -and $ZipPassword -ne '') {
		$Body.add("password",$ZipPassword)
	}

	if ($EncryptData -ne $null) {
		$Body.add("encryptData",$EncryptData)
	} else {
		$Body.add("encryptData", $false)
	}

	Invoke-anynodeDownload -Address $Address -Port $Port -URITail /backup -Body ($Body | ConvertTo-Json) -Filename $Filename -Method Post -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Parameters $Parameters -Username $Username -Password $Password
}
# SIG # Begin signature block
# MIIt5QYJKoZIhvcNAQcCoIIt1jCCLdICAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUi8uvZ8V++T6horqrJBTyLb1x
# YsuggiceMIIFcjCCA1qgAwIBAgIQdlP+rHVGSJP15ddKSDpO+DANBgkqhkiG9w0B
# AQwFADBTMQswCQYDVQQGEwJCRTEZMBcGA1UEChMQR2xvYmFsU2lnbiBudi1zYTEp
# MCcGA1UEAxMgR2xvYmFsU2lnbiBDb2RlIFNpZ25pbmcgUm9vdCBSNDUwHhcNMjAw
# MzE4MDAwMDAwWhcNNDUwMzE4MDAwMDAwWjBTMQswCQYDVQQGEwJCRTEZMBcGA1UE
# ChMQR2xvYmFsU2lnbiBudi1zYTEpMCcGA1UEAxMgR2xvYmFsU2lnbiBDb2RlIFNp
# Z25pbmcgUm9vdCBSNDUwggIiMA0GCSqGSIb3DQEBAQUAA4ICDwAwggIKAoICAQC2
# LcUw3Xroq5A9A3KwOkuZFmGy5f+lZx03HOV+7JODqoT1o0ObmEWKuGNXXZsAiAQl
# 6fhokkuC2EvJSgPzqH9qj4phJ72hRND99T8iwqNPkY2zBbIogpFd+1mIBQuXBsKY
# +CynMyTuUDpBzPCgsHsdTdKoWDiW6d/5G5G7ixAs0sdDHaIJdKGAr3vmMwoMWWuO
# vPSrWpd7f65V+4TwgP6ETNfiur3EdaFvvWEQdESymAfidKv/aNxsJj7pH+XgBIet
# MNMMjQN8VbgWcFwkeCAl62dniKu6TjSYa3AR3jjK1L6hwJzh3x4CAdg74WdDhLbP
# /HS3L4Sjv7oJNz1nbLFFXBlhq0GD9awd63cNRkdzzr+9lZXtnSuIEP76WOinV+Gz
# z6ha6QclmxLEnoByPZPcjJTfO0TmJoD80sMD8IwM0kXWLuePmJ7mBO5Cbmd+QhZx
# YucE+WDGZKG2nIEhTivGbWiUhsaZdHNnMXqR8tSMeW58prt+Rm9NxYUSK8+aIkQI
# qIU3zgdhVwYXEiTAxDFzoZg1V0d+EDpF2S2kUZCYqaAHN8RlGqocaxZ396eX7D8Z
# MJlvMfvqQLLn0sT6ydDwUHZ0WfqNbRcyvvjpfgP054d1mtRKkSyFAxMCK0KA8olq
# Ns/ITKDOnvjLja0Wp9Pe1ZsYp8aSOvGCY/EuDiRk3wIDAQABo0IwQDAOBgNVHQ8B
# Af8EBAMCAYYwDwYDVR0TAQH/BAUwAwEB/zAdBgNVHQ4EFgQUHwC/RoAK/Hg5t6W0
# Q9lWULvOljswDQYJKoZIhvcNAQEMBQADggIBAF4runSXNERfdkgoQIST7gFu6aGz
# 1oAl5nvkvAmRPQ/8dq3X1DAgu49g0JHWHPKc73gaK5QyAsEkllJSAtDz0fzymzlu
# meEfjkNBfZoeW8ldmoT8JuaH83RyJq2kG9k9O2pSoDwJHi8ee7MztEXH96yxr5Ng
# rXauuLIVeOuDauv/20arJOXuAvqQH1nAL13Wt12kXBC3clP4QU7M+ngaJUrK/oVi
# Q2HDtDeqgdL01joPvY1ZfjBH3itr5yFQM1/UZ5vUuGefPCeZA/+FQ45zEsogzehh
# 1bFm3BfWOW0P288jN6GCiU4caz/WoM2qB50+Qiaq1wzu+ke/GlJ+0XWB08mKYhdt
# T4igIaAmPq9t2WIwH+mYKK5ujdWOTHJmk4CNKuNVx2BnkEJWXCJRD7PcTjnuTd3Z
# HXgQVDtu0JdvA7UesiNzxhKymmTQ/JWFJKj/36Gw3JFArt8JM6u53ZK38cyRdDtp
# 62eXG5C/58egb3G7V7+3j1rtekBqFs2AhC0v4QLUJJRDsxX8DCsb/XFv/Mu8dRc6
# XoPSybMvG9WcjX9U/n5+5Fajh6ed4VlSlEGPbVu+hpWa/xp23UDSUUpwtB8zYyN3
# P+wnHlnkCIftNIJKDz/+oB3B9WdzRYZ49Kop6SeHxhnbxhMUwzlJh02gl+BlE/Wd
# d1bp2rNYxzrywM2CMIIFjTCCBHWgAwIBAgIQDpsYjvnQLefv21DiCEAYWjANBgkq
# hkiG9w0BAQwFADBlMQswCQYDVQQGEwJVUzEVMBMGA1UEChMMRGlnaUNlcnQgSW5j
# MRkwFwYDVQQLExB3d3cuZGlnaWNlcnQuY29tMSQwIgYDVQQDExtEaWdpQ2VydCBB
# c3N1cmVkIElEIFJvb3QgQ0EwHhcNMjIwODAxMDAwMDAwWhcNMzExMTA5MjM1OTU5
# WjBiMQswCQYDVQQGEwJVUzEVMBMGA1UEChMMRGlnaUNlcnQgSW5jMRkwFwYDVQQL
# ExB3d3cuZGlnaWNlcnQuY29tMSEwHwYDVQQDExhEaWdpQ2VydCBUcnVzdGVkIFJv
# b3QgRzQwggIiMA0GCSqGSIb3DQEBAQUAA4ICDwAwggIKAoICAQC/5pBzaN675F1K
# PDAiMGkz7MKnJS7JIT3yithZwuEppz1Yq3aaza57G4QNxDAf8xukOBbrVsaXbR2r
# snnyyhHS5F/WBTxSD1Ifxp4VpX6+n6lXFllVcq9ok3DCsrp1mWpzMpTREEQQLt+C
# 8weE5nQ7bXHiLQwb7iDVySAdYyktzuxeTsiT+CFhmzTrBcZe7FsavOvJz82sNEBf
# sXpm7nfISKhmV1efVFiODCu3T6cw2Vbuyntd463JT17lNecxy9qTXtyOj4DatpGY
# QJB5w3jHtrHEtWoYOAMQjdjUN6QuBX2I9YI+EJFwq1WCQTLX2wRzKm6RAXwhTNS8
# rhsDdV14Ztk6MUSaM0C/CNdaSaTC5qmgZ92kJ7yhTzm1EVgX9yRcRo9k98FpiHaY
# dj1ZXUJ2h4mXaXpI8OCiEhtmmnTK3kse5w5jrubU75KSOp493ADkRSWJtppEGSt+
# wJS00mFt6zPZxd9LBADMfRyVw4/3IbKyEbe7f/LVjHAsQWCqsWMYRJUadmJ+9oCw
# ++hkpjPRiQfhvbfmQ6QYuKZ3AeEPlAwhHbJUKSWJbOUOUlFHdL4mrLZBdd56rF+N
# P8m800ERElvlEFDrMcXKchYiCd98THU/Y+whX8QgUWtvsauGi0/C1kVfnSD8oR7F
# wI+isX4KJpn15GkvmB0t9dmpsh3lGwIDAQABo4IBOjCCATYwDwYDVR0TAQH/BAUw
# AwEB/zAdBgNVHQ4EFgQU7NfjgtJxXWRM3y5nP+e6mK4cD08wHwYDVR0jBBgwFoAU
# Reuir/SSy4IxLVGLp6chnfNtyA8wDgYDVR0PAQH/BAQDAgGGMHkGCCsGAQUFBwEB
# BG0wazAkBggrBgEFBQcwAYYYaHR0cDovL29jc3AuZGlnaWNlcnQuY29tMEMGCCsG
# AQUFBzAChjdodHRwOi8vY2FjZXJ0cy5kaWdpY2VydC5jb20vRGlnaUNlcnRBc3N1
# cmVkSURSb290Q0EuY3J0MEUGA1UdHwQ+MDwwOqA4oDaGNGh0dHA6Ly9jcmwzLmRp
# Z2ljZXJ0LmNvbS9EaWdpQ2VydEFzc3VyZWRJRFJvb3RDQS5jcmwwEQYDVR0gBAow
# CDAGBgRVHSAAMA0GCSqGSIb3DQEBDAUAA4IBAQBwoL9DXFXnOF+go3QbPbYW1/e/
# Vwe9mqyhhyzshV6pGrsi+IcaaVQi7aSId229GhT0E0p6Ly23OO/0/4C5+KH38nLe
# JLxSA8hO0Cre+i1Wz/n096wwepqLsl7Uz9FDRJtDIeuWcqFItJnLnU+nBgMTdydE
# 1Od/6Fmo8L8vC6bp8jQ87PcDx4eo0kxAGTVGamlUsLihVo7spNU96LHc/RzY9Hda
# XFSMb++hUD38dglohJ9vytsgjTVgHAIDyyCwrFigDkBjxZgiwbJZ9VVrzyerbHbO
# byMt9H5xaiNrIv8SuFQtJ37YOtnwtoeW/VvRXKwYw02fc7cBqZ9Xql4o4rmUMIIG
# rjCCBJagAwIBAgIQBzY3tyRUfNhHrP0oZipeWzANBgkqhkiG9w0BAQsFADBiMQsw
# CQYDVQQGEwJVUzEVMBMGA1UEChMMRGlnaUNlcnQgSW5jMRkwFwYDVQQLExB3d3cu
# ZGlnaWNlcnQuY29tMSEwHwYDVQQDExhEaWdpQ2VydCBUcnVzdGVkIFJvb3QgRzQw
# HhcNMjIwMzIzMDAwMDAwWhcNMzcwMzIyMjM1OTU5WjBjMQswCQYDVQQGEwJVUzEX
# MBUGA1UEChMORGlnaUNlcnQsIEluYy4xOzA5BgNVBAMTMkRpZ2lDZXJ0IFRydXN0
# ZWQgRzQgUlNBNDA5NiBTSEEyNTYgVGltZVN0YW1waW5nIENBMIICIjANBgkqhkiG
# 9w0BAQEFAAOCAg8AMIICCgKCAgEAxoY1BkmzwT1ySVFVxyUDxPKRN6mXUaHW0oPR
# nkyibaCwzIP5WvYRoUQVQl+kiPNo+n3znIkLf50fng8zH1ATCyZzlm34V6gCff1D
# tITaEfFzsbPuK4CEiiIY3+vaPcQXf6sZKz5C3GeO6lE98NZW1OcoLevTsbV15x8G
# ZY2UKdPZ7Gnf2ZCHRgB720RBidx8ald68Dd5n12sy+iEZLRS8nZH92GDGd1ftFQL
# IWhuNyG7QKxfst5Kfc71ORJn7w6lY2zkpsUdzTYNXNXmG6jBZHRAp8ByxbpOH7G1
# WE15/tePc5OsLDnipUjW8LAxE6lXKZYnLvWHpo9OdhVVJnCYJn+gGkcgQ+NDY4B7
# dW4nJZCYOjgRs/b2nuY7W+yB3iIU2YIqx5K/oN7jPqJz+ucfWmyU8lKVEStYdEAo
# q3NDzt9KoRxrOMUp88qqlnNCaJ+2RrOdOqPVA+C/8KI8ykLcGEh/FDTP0kyr75s9
# /g64ZCr6dSgkQe1CvwWcZklSUPRR8zZJTYsg0ixXNXkrqPNFYLwjjVj33GHek/45
# wPmyMKVM1+mYSlg+0wOI/rOP015LdhJRk8mMDDtbiiKowSYI+RQQEgN9XyO7ZONj
# 4KbhPvbCdLI/Hgl27KtdRnXiYKNYCQEoAA6EVO7O6V3IXjASvUaetdN2udIOa5kM
# 0jO0zbECAwEAAaOCAV0wggFZMBIGA1UdEwEB/wQIMAYBAf8CAQAwHQYDVR0OBBYE
# FLoW2W1NhS9zKXaaL3WMaiCPnshvMB8GA1UdIwQYMBaAFOzX44LScV1kTN8uZz/n
# upiuHA9PMA4GA1UdDwEB/wQEAwIBhjATBgNVHSUEDDAKBggrBgEFBQcDCDB3Bggr
# BgEFBQcBAQRrMGkwJAYIKwYBBQUHMAGGGGh0dHA6Ly9vY3NwLmRpZ2ljZXJ0LmNv
# bTBBBggrBgEFBQcwAoY1aHR0cDovL2NhY2VydHMuZGlnaWNlcnQuY29tL0RpZ2lD
# ZXJ0VHJ1c3RlZFJvb3RHNC5jcnQwQwYDVR0fBDwwOjA4oDagNIYyaHR0cDovL2Ny
# bDMuZGlnaWNlcnQuY29tL0RpZ2lDZXJ0VHJ1c3RlZFJvb3RHNC5jcmwwIAYDVR0g
# BBkwFzAIBgZngQwBBAIwCwYJYIZIAYb9bAcBMA0GCSqGSIb3DQEBCwUAA4ICAQB9
# WY7Ak7ZvmKlEIgF+ZtbYIULhsBguEE0TzzBTzr8Y+8dQXeJLKftwig2qKWn8acHP
# HQfpPmDI2AvlXFvXbYf6hCAlNDFnzbYSlm/EUExiHQwIgqgWvalWzxVzjQEiJc6V
# aT9Hd/tydBTX/6tPiix6q4XNQ1/tYLaqT5Fmniye4Iqs5f2MvGQmh2ySvZ180HAK
# fO+ovHVPulr3qRCyXen/KFSJ8NWKcXZl2szwcqMj+sAngkSumScbqyQeJsG33irr
# 9p6xeZmBo1aGqwpFyd/EjaDnmPv7pp1yr8THwcFqcdnGE4AJxLafzYeHJLtPo0m5
# d2aR8XKc6UsCUqc3fpNTrDsdCEkPlM05et3/JWOZJyw9P2un8WbDQc1PtkCbISFA
# 0LcTJM3cHXg65J6t5TRxktcma+Q4c6umAU+9Pzt4rUyt+8SVe+0KXzM5h0F4ejjp
# nOHdI/0dKNPH+ejxmF/7K9h+8kaddSweJywm228Vex4Ziza4k9Tm8heZWcpw8De/
# mADfIBZPJ/tgZxahZrrdVcA6KYawmKAr7ZVBtzrVFZgxtGIJDwq9gdkT/r+k0fNX
# 2bwE+oLeMt8EifAAzV3C+dAjfwAL5HYCJtnwZXZCpimHCUcr5n8apIUP/JiW9lVU
# Kx+A+sDyDivl1vupL0QVSucTDh3bNzgaoSv27dZ8/DCCBsIwggSqoAMCAQICEAVE
# r/OUnQg5pr/bP1/lYRYwDQYJKoZIhvcNAQELBQAwYzELMAkGA1UEBhMCVVMxFzAV
# BgNVBAoTDkRpZ2lDZXJ0LCBJbmMuMTswOQYDVQQDEzJEaWdpQ2VydCBUcnVzdGVk
# IEc0IFJTQTQwOTYgU0hBMjU2IFRpbWVTdGFtcGluZyBDQTAeFw0yMzA3MTQwMDAw
# MDBaFw0zNDEwMTMyMzU5NTlaMEgxCzAJBgNVBAYTAlVTMRcwFQYDVQQKEw5EaWdp
# Q2VydCwgSW5jLjEgMB4GA1UEAxMXRGlnaUNlcnQgVGltZXN0YW1wIDIwMjMwggIi
# MA0GCSqGSIb3DQEBAQUAA4ICDwAwggIKAoICAQCjU0WHHYOOW6w+VLMj4M+f1+XS
# 512hDgncL0ijl3o7Kpxn3GIVWMGpkxGnzaqyat0QKYoeYmNp01icNXG/OpfrlFCP
# HCDqx5o7L5Zm42nnaf5bw9YrIBzBl5S0pVCB8s/LB6YwaMqDQtr8fwkklKSCGtpq
# utg7yl3eGRiF+0XqDWFsnf5xXsQGmjzwxS55DxtmUuPI1j5f2kPThPXQx/ZILV5F
# dZZ1/t0QoRuDwbjmUpW1R9d4KTlr4HhZl+NEK0rVlc7vCBfqgmRN/yPjyobutKQh
# ZHDr1eWg2mOzLukF7qr2JPUdvJscsrdf3/Dudn0xmWVHVZ1KJC+sK5e+n+T9e3M+
# Mu5SNPvUu+vUoCw0m+PebmQZBzcBkQ8ctVHNqkxmg4hoYru8QRt4GW3k2Q/gWEH7
# 2LEs4VGvtK0VBhTqYggT02kefGRNnQ/fztFejKqrUBXJs8q818Q7aESjpTtC/XN9
# 7t0K/3k0EH6mXApYTAA+hWl1x4Nk1nXNjxJ2VqUk+tfEayG66B80mC866msBsPf7
# Kobse1I4qZgJoXGybHGvPrhvltXhEBP+YUcKjP7wtsfVx95sJPC/QoLKoHE9nJKT
# BLRpcCcNT7e1NtHJXwikcKPsCvERLmTgyyIryvEoEyFJUX4GZtM7vvrrkTjYUQfK
# lLfiUKHzOtOKg8tAewIDAQABo4IBizCCAYcwDgYDVR0PAQH/BAQDAgeAMAwGA1Ud
# EwEB/wQCMAAwFgYDVR0lAQH/BAwwCgYIKwYBBQUHAwgwIAYDVR0gBBkwFzAIBgZn
# gQwBBAIwCwYJYIZIAYb9bAcBMB8GA1UdIwQYMBaAFLoW2W1NhS9zKXaaL3WMaiCP
# nshvMB0GA1UdDgQWBBSltu8T5+/N0GSh1VapZTGj3tXjSTBaBgNVHR8EUzBRME+g
# TaBLhklodHRwOi8vY3JsMy5kaWdpY2VydC5jb20vRGlnaUNlcnRUcnVzdGVkRzRS
# U0E0MDk2U0hBMjU2VGltZVN0YW1waW5nQ0EuY3JsMIGQBggrBgEFBQcBAQSBgzCB
# gDAkBggrBgEFBQcwAYYYaHR0cDovL29jc3AuZGlnaWNlcnQuY29tMFgGCCsGAQUF
# BzAChkxodHRwOi8vY2FjZXJ0cy5kaWdpY2VydC5jb20vRGlnaUNlcnRUcnVzdGVk
# RzRSU0E0MDk2U0hBMjU2VGltZVN0YW1waW5nQ0EuY3J0MA0GCSqGSIb3DQEBCwUA
# A4ICAQCBGtbeoKm1mBe8cI1PijxonNgl/8ss5M3qXSKS7IwiAqm4z4Co2efjxe0m
# gopxLxjdTrbebNfhYJwr7e09SI64a7p8Xb3CYTdoSXej65CqEtcnhfOOHpLawkA4
# n13IoC4leCWdKgV6hCmYtld5j9smViuw86e9NwzYmHZPVrlSwradOKmB521BXIxp
# 0bkrxMZ7z5z6eOKTGnaiaXXTUOREEr4gDZ6pRND45Ul3CFohxbTPmJUaVLq5vMFp
# GbrPFvKDNzRusEEm3d5al08zjdSNd311RaGlWCZqA0Xe2VC1UIyvVr1MxeFGxSjT
# redDAHDezJieGYkD6tSRN+9NUvPJYCHEVkft2hFLjDLDiOZY4rbbPvlfsELWj+MX
# kdGqwFXjhr+sJyxB0JozSqg21Llyln6XeThIX8rC3D0y33XWNmdaifj2p8flTzU8
# AL2+nCpseQHc2kTmOt44OwdeOVj0fHMxVaCAEcsUDH6uvP6k63llqmjWIso765qC
# NVcoFstp8jKastLYOrixRoZruhf9xHdsFWyuq69zOuhJRrfVf8y2OMDY7Bz1tqG4
# QyzfTkx9HmhwwHcK1ALgXGC7KP845VJa1qwXIiNO9OzTF/tQa/8Hdx9xl0RBybhG
# 02wyfFgvZ0dl5Rtztpn5aywGRu9BHvDwX+Db2a2QgESvgBBBijCCBugwggTQoAMC
# AQICEHe9DgW3WQu2HUdhUx4/de0wDQYJKoZIhvcNAQELBQAwUzELMAkGA1UEBhMC
# QkUxGTAXBgNVBAoTEEdsb2JhbFNpZ24gbnYtc2ExKTAnBgNVBAMTIEdsb2JhbFNp
# Z24gQ29kZSBTaWduaW5nIFJvb3QgUjQ1MB4XDTIwMDcyODAwMDAwMFoXDTMwMDcy
# ODAwMDAwMFowXDELMAkGA1UEBhMCQkUxGTAXBgNVBAoTEEdsb2JhbFNpZ24gbnYt
# c2ExMjAwBgNVBAMTKUdsb2JhbFNpZ24gR0NDIFI0NSBFViBDb2RlU2lnbmluZyBD
# QSAyMDIwMIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEAyyDvlx65ATJD
# oFupiiP9IF6uOBKLyizU/0HYGlXUGVO3/aMX53o5XMD3zhGj+aXtAfq1upPvr5Pc
# +OKzGUyDsEpEUAR4hBBqpNaWkI6B+HyrL7WjVzPSWHuUDm0PpZEmKrODT3Kxintk
# ktDwtFVflgsR5Zq1LLIRzyUbfVErmB9Jo1/4E541uAMC2qQTL4VK78QvcA7B1Mwz
# Euy9QJXTEcrmzbMFnMhT61LXeExRAZKC3hPzB450uoSAn9KkFQ7or+v3ifbfcfDR
# vqeyQTMgdcyx1e0dBxnE6yZ38qttF5NJqbfmw5CcxrjszMl7ml7FxSSTY29+EIth
# z5hVoySiiDby+Z++ky6yBp8mwAwBVhLhsoqfDh7cmIsuz9riiTSmHyagqK54beyh
# iBU8wurut9itYaWvcDaieY7cDXPA8eQsq5TsWAY5NkjWO1roIs50Dq8s8RXa0bSV
# 6KzVSW3lr92ba2MgXY5+O7JD2GI6lOXNtJizNxkkEnJzqwSwCdyF5tQiBO9AKh0u
# bcdp0263AWwN4JenFuYmi4j3A0SGX2JnTLWnN6hV3AM2jG7PbTYm8Q6PsD1xwOEy
# p4LktjICMjB8tZPIIf08iOZpY/judcmLwqvvujr96V6/thHxvvA9yjI+bn3eD36b
# lcQSh+cauE7uLMHfoWXoJIPJKsL9uVMCAwEAAaOCAa0wggGpMA4GA1UdDwEB/wQE
# AwIBhjATBgNVHSUEDDAKBggrBgEFBQcDAzASBgNVHRMBAf8ECDAGAQH/AgEAMB0G
# A1UdDgQWBBQlndD8WQmGY8Xs87ETO1ccA5I2ETAfBgNVHSMEGDAWgBQfAL9GgAr8
# eDm3pbRD2VZQu86WOzCBkwYIKwYBBQUHAQEEgYYwgYMwOQYIKwYBBQUHMAGGLWh0
# dHA6Ly9vY3NwLmdsb2JhbHNpZ24uY29tL2NvZGVzaWduaW5ncm9vdHI0NTBGBggr
# BgEFBQcwAoY6aHR0cDovL3NlY3VyZS5nbG9iYWxzaWduLmNvbS9jYWNlcnQvY29k
# ZXNpZ25pbmdyb290cjQ1LmNydDBBBgNVHR8EOjA4MDagNKAyhjBodHRwOi8vY3Js
# Lmdsb2JhbHNpZ24uY29tL2NvZGVzaWduaW5ncm9vdHI0NS5jcmwwVQYDVR0gBE4w
# TDBBBgkrBgEEAaAyAQIwNDAyBggrBgEFBQcCARYmaHR0cHM6Ly93d3cuZ2xvYmFs
# c2lnbi5jb20vcmVwb3NpdG9yeS8wBwYFZ4EMAQMwDQYJKoZIhvcNAQELBQADggIB
# ACV1oAnJObq3oTmJLxifq9brHUvolHwNB2ibHJ3vcbYXamsCT7M/hkWHzGWbTONY
# BgIiZtVhAsVjj9Si8bZeJQt3lunNcUAziCns7vOibbxNtT4GS8lzM8oIFC09TOiw
# unWmdC2kWDpsE0n4pRUKFJaFsWpoNCVCr5ZW9BD6JH3xK3LBFuFr6+apmMc+WvTQ
# GJ39dJeGd0YqPSN9KHOKru8rG5q/bFOnFJ48h3HAXo7I+9MqkjPqV01eB17KwRis
# gS0aIfpuz5dhe99xejrKY/fVMEQ3Mv67Q4XcuvymyjMZK3dt28sF8H5fdS6itr81
# qjZjyc5k2b38vCzzSVYAyBIrxie7N69X78TPHinE9OItziphz1ft9QpA4vUY1h7p
# kC/K04dfk4pIGhEd5TeFny5mYppegU6VrFVXQ9xTiyV+PGEPigu69T+m1473BFZe
# Ibuf12pxgL+W3nID2NgiK/MnFk846FFADK6S7749ffeAxkw2V4SVp4QVSDAOUicI
# jY6ivSLHGcmmyg6oejbbarphXxEklaTijmjuGalJmV7QtDS91vlAxxCXMVI5NSkR
# hyTTxPupY8t3SNX6Yvwk4AR6TtDkbt7OnjhQJvQhcWXXCSXUyQcAerjH83foxdTi
# VdDTHvZ/UuJJjbkRcgyIRCYzZgFE3+QzDiHeYolIB9r1MIIHrzCCBZegAwIBAgIM
# H7b9bdAxkf8/itS3MA0GCSqGSIb3DQEBCwUAMFwxCzAJBgNVBAYTAkJFMRkwFwYD
# VQQKExBHbG9iYWxTaWduIG52LXNhMTIwMAYDVQQDEylHbG9iYWxTaWduIEdDQyBS
# NDUgRVYgQ29kZVNpZ25pbmcgQ0EgMjAyMDAeFw0yMjAzMTQxNjAyNTZaFw0yNTA0
# MjQwOTA3MjBaMIIBEzEdMBsGA1UEDwwUUHJpdmF0ZSBPcmdhbml6YXRpb24xEzAR
# BgNVBAUTCkhSQiAxMDA1MjExEzARBgsrBgEEAYI3PAIBAxMCREUxHjAcBgsrBgEE
# AYI3PAIBAhMNTmllZGVyc2FjaHNlbjEdMBsGCysGAQQBgjc8AgEBEwxCcmF1bnNj
# aHdlaWcxCzAJBgNVBAYTAkRFMRYwFAYDVQQIEw1OaWVkZXJzYWNoc2VuMRIwEAYD
# VQQHEwlXb2xmc2J1cmcxHDAaBgNVBAkTE01heC12b24tTGF1ZS1XZWcgMTkxGDAW
# BgNVBAoTD1RFLVNZU1RFTVMgR21iSDEYMBYGA1UEAxMPVEUtU1lTVEVNUyBHbWJI
# MIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEAsNpWT+JEuh43teGS+r3W
# 5zFgXA4a42GqnmEOIkUp+UPqC+6k9++zXUFIZjlCVluKJ3Xv4NDyrDu26i6/P96J
# iLzBJOX9uZItduUBL7o3lANnjom7FEVpmJJGf+mbxvB04tdSvhf9IZsxpPZEaOth
# ZAQaWDTsg0LzZQeHAxaOueoVOG9CHQRsJAwFX3HhaVHSd8atmi7kDXQAkpwDn6+2
# W1OT/IQKUAEr2vxysctY5Bx4PCX/ZxRk1J/D5WyapcHOZyMMCtMXUTFse4nvLxMS
# j6ha94PCdCZC4bOijYguV3JAYuCsFH99s36cLUpLxX7pnH0I6KIEttXIyER+WAbI
# oNr2UDgd6RZJ+XsU4WE7w5YguOiOwiBHn3/guBorvFOFaVPHVu8N9ZqjfInvvAdS
# fqT5rBNryGWvUrB2SPd0y21DxC4DKph0IcwVPCJiP/5dtAS3aXZTePco8+W/PNNo
# 0uu4wwLfGgZwJSQFwTtA7vaKmhGPgT8mBfhtB2UjMG+8XT7gr3ov/+5hg+Ru/u2J
# FI7N1O/jr+1gs6JlK4YF8ETFxXC+nETwykd49OugjVyRQEhQjnHM48gvvFq/uL2K
# IVbf6gojJHW93KYiGPMQ2gcXcfnNcFu2Ro69ve58OQxNxYSKIi/sXCRKoNC65HUs
# 7i+riSNPNVBIzaqPgHU6iXECAwEAAaOCAbYwggGyMA4GA1UdDwEB/wQEAwIHgDCB
# nwYIKwYBBQUHAQEEgZIwgY8wTAYIKwYBBQUHMAKGQGh0dHA6Ly9zZWN1cmUuZ2xv
# YmFsc2lnbi5jb20vY2FjZXJ0L2dzZ2NjcjQ1ZXZjb2Rlc2lnbmNhMjAyMC5jcnQw
# PwYIKwYBBQUHMAGGM2h0dHA6Ly9vY3NwLmdsb2JhbHNpZ24uY29tL2dzZ2NjcjQ1
# ZXZjb2Rlc2lnbmNhMjAyMDBVBgNVHSAETjBMMEEGCSsGAQQBoDIBAjA0MDIGCCsG
# AQUFBwIBFiZodHRwczovL3d3dy5nbG9iYWxzaWduLmNvbS9yZXBvc2l0b3J5LzAH
# BgVngQwBAzAJBgNVHRMEAjAAMEcGA1UdHwRAMD4wPKA6oDiGNmh0dHA6Ly9jcmwu
# Z2xvYmFsc2lnbi5jb20vZ3NnY2NyNDVldmNvZGVzaWduY2EyMDIwLmNybDATBgNV
# HSUEDDAKBggrBgEFBQcDAzAfBgNVHSMEGDAWgBQlndD8WQmGY8Xs87ETO1ccA5I2
# ETAdBgNVHQ4EFgQUtGt3xHpCCOM3zHPw8Rlox5B+tvQwDQYJKoZIhvcNAQELBQAD
# ggIBAHwLtY4rR+w31n+vChCCzr6LIw/iefwzwLQ33L0UOm9alIEnxcISt4Fna0WJ
# Nd3p8f8AUQ9rv0qbO8QNyS9kPd34rN/K8SqCj3gtEyBBTaGWTz7SAnynn3Wid1bw
# FMzlkebJ7m9ug1eOhnKJudeiadKLetz32Miaagr4pEM0r7zeKanW4voDLAHib/fA
# NEEBUq2j9f3WPnw2shLhG4PRVIqpnBtqf/nR3hhEldibZx4xuI+OEgjnC49Rz4cy
# ZuZLw1SO6R4vwx9cHtnZC5kZPE+awrj72ic0lQxbvK9HJTJBUOJuYRjVA86uxjPG
# wvaR8XdeX0Xw1ytdJd4X7iCvj2cqO92iV/yq+Y1mLapJi5POezf6LdDcLUKF4vpJ
# w0bFGz9A5ZwnXhNQ2HTFOIg7hjoq+dY82dfnh+fHVfFNU2OZZ6CxEXad+Gaw4grA
# LnLHbitzbue0FXPW3yoWWVsjLhHWZFhfTOpXgziZB0slTI1kXydXfhfcFoFladlc
# BA3dQnfieH+fuVFc2cYSfS1UJu5juDjB3UjRZYTAe+cKq/2IGlK/eiyGjxxKkPCc
# hrlp70FyDlSgAyl+NUYNCAGAgHZlS3x7usATC8rlfKffuXPHzCepG0/Lpvj0b6PM
# p5XuRnRpvlk0j5CbwZVguLIsvFVHzd9z/gUsvhqHSvcP08+iMYIGMTCCBi0CAQEw
# bDBcMQswCQYDVQQGEwJCRTEZMBcGA1UEChMQR2xvYmFsU2lnbiBudi1zYTEyMDAG
# A1UEAxMpR2xvYmFsU2lnbiBHQ0MgUjQ1IEVWIENvZGVTaWduaW5nIENBIDIwMjAC
# DB+2/W3QMZH/P4rUtzAJBgUrDgMCGgUAoHgwGAYKKwYBBAGCNwIBDDEKMAigAoAA
# oQKAADAZBgkqhkiG9w0BCQMxDAYKKwYBBAGCNwIBBDAcBgorBgEEAYI3AgELMQ4w
# DAYKKwYBBAGCNwIBFTAjBgkqhkiG9w0BCQQxFgQUixQrtqzlq8s27EviBRgb1CER
# KnQwDQYJKoZIhvcNAQEBBQAEggIAJm8uNilBYhb1H/ZIWvF2NTydbyvE+Y1LxiiG
# MITXjADBftdwc7AybS0ThldnG+CT78UgpXxEfn5JM0la7rkXYrK5dZC/KUV6n3ku
# OYUt3o31ADqXS1Se6SHBDqqppv5F0+FvCeq1LUFoBuQGuJqFvMx9aegXIYONSDCO
# CgGg0f+1NEzAlsW0LaZpWDjh8lhHzQzdI9RLzeVIdgyRWl8BuXFEZjgd97QaFf8p
# XwnysBpiVoNpxHiCHuT4lPiQYkeruKQK8sPDXRbFlauNRIk0YHlPOzppo/tEs/kp
# AmoZVBPrnYmz1tt9C2WdHV1PhHvrmO1UPDUXldKE6RhdLMlax/QWlwN84Igv2qLx
# 8LI+NuiyhkJZ8bb0ZwtEhd50ut50TKb+WyRi11Z6U4PLw4sjI+soyWO4SrcesVQU
# Tb5074U9bCli71xHsccNhD2WDWatCgEs62Tvptti2bgoJ3EjvyAZz9syKrBD8uJH
# NavzCJxyEpzHDptTRyzjVJrVrdBKVsnkt5nLIY+YNhol1Wt4ZFUaEVoyJWXcsNe6
# rVDtzWfBNPnNNqgkQxHVT/fTCb8Vz5iSY8NPOf+Z+L+LenVQ5TRjFEFgViohj/4A
# 9CQiFSl/kjyODjLdh+gUfLWgg6U3kUEnQws+BsWIlf5hBcfueT4t+0EsalTgbyvF
# FEOBJQGhggMgMIIDHAYJKoZIhvcNAQkGMYIDDTCCAwkCAQEwdzBjMQswCQYDVQQG
# EwJVUzEXMBUGA1UEChMORGlnaUNlcnQsIEluYy4xOzA5BgNVBAMTMkRpZ2lDZXJ0
# IFRydXN0ZWQgRzQgUlNBNDA5NiBTSEEyNTYgVGltZVN0YW1waW5nIENBAhAFRK/z
# lJ0IOaa/2z9f5WEWMA0GCWCGSAFlAwQCAQUAoGkwGAYJKoZIhvcNAQkDMQsGCSqG
# SIb3DQEHATAcBgkqhkiG9w0BCQUxDxcNMjQwMTE2MDkyNTU0WjAvBgkqhkiG9w0B
# CQQxIgQgrdvpTkoyFHtLrTrqnldsszb45Qg4TD4jZDbQpR2SsWgwDQYJKoZIhvcN
# AQEBBQAEggIAfQXZ0i3hMfA23ngwROQ+Qa7+2WMaIAgkO+iu8BZsnCi3qjVFroF8
# rTMj7auYP1Vy2qS6NnGqQezwPI+wrHLN5pN9NljhlucUgra7+RdZ2gg3oVgJhxYH
# smHyOVB9N13B42kC6uI1pLHP1G9+g07nBweGEtLlJ1GFREbECEVaqYmD2bzSioyw
# 1pU3ueQ5iwgz9n+SpYBI/M183cZH+JyELuavvKJ+b9BQNKVIPW25gkLYIEDw1HEt
# R1+HZKi+EEv3G0YBIiCo9Z1TJcRKPcbdvPJGlxvRbB0RX9y24eS2wJeNsawW5/xg
# 2SHuGu/lY567TsJp/gCflX56gJNR/+rxanbwNWw30ImyY5FuFLAYfY+dHFpuR92h
# gsXaIeDPjdPQGQicTmcsH4+/MmriJaYEAM7hRqJ6tXBACIuRmgHoZa0ngejRZzXK
# 5Cnt8rbhznh26cX1XdCjBbqcL6zDi3uk/6ClKHohT3S7DXCB81yeOHCWYRQpZ+fg
# yBzv/+rmHmVFwxOqsR/RxBeTPiESXiDxaCxm4bmV66FNYXI1Wf2cimo0pnPZBrSC
# ntAyg3y4iD/7g/qz818A+iBk2AZSPyjSK3GoW4XsQdJ2wuZqe2W7SN4WM445SCWR
# +FLUx6obHBsgu6sR7cB+gpIUZT29W4tppt5bVBYRINiYhcBBf9rAk0k=
# SIG # End signature block
